#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from functools import partial
from threading import Thread
from time import sleep
from typing import Any, List, Optional, Callable, Tuple
import pickle

from PySide2.QtWidgets import (
    QWidget,
    QScrollArea,
    QGridLayout,
    QVBoxLayout,
    QStyle,
)
from PySide2.QtCore import Qt, QObject, QThread
from PySide2.QtGui import QCloseEvent

from scine_database import Manager, Model
from scine_utilities import ValueCollection

from scine_chemoton.filters.aggregate_filters import \
    AggregateFilter
from scine_chemoton.filters.reactive_site_filters import \
    ReactiveSiteFilter
from scine_chemoton.filters.further_exploration_filters import \
    FurtherExplorationFilter
from scine_chemoton.filters.elementary_step_filters import ElementaryStepFilter
from scine_chemoton.filters.reaction_filters import ReactionFilter
from scine_chemoton.gears.network_refinement.enabling import (
    AggregateEnabling, ReactionEnabling, EnableCalculationResults
)
from scine_chemoton.gears.network_refinement.disabling import ReactionDisabling, StepDisabling
from scine_chemoton.filters.structure_filters import StructureFilter
from scine_chemoton.utilities.datastructure_transfer import make_picklable

from scine_heron.containers.tab_widget import TabWidget
from scine_heron.containers.start_stop_widget import StartStopWidget
from scine_heron.chemoton.class_searcher import ChemotonClassSearcher
from scine_heron.settings.class_options_widget import ModelOptionsWidget, GeneralSettingsWidget
from scine_heron.settings.docstring_parser import DocStringParser
from scine_heron.chemoton.filter_builder import FilterBuilder
from scine_heron.io.file_browser_popup import get_load_file_name, get_save_file_name
from scine_heron.toolbar.io_toolbar import HeronToolBar
from scine_heron.utilities import write_error_message, write_info_message, clear_status_bar
import scine_heron.io.json_pickle_wrap as json_wrap


class ChemotonWidgetContainer(QWidget):
    """
    Container Widget for StartStopWidgets in a grid view.
    This represents the whole Chemoton tab.
    """

    def __init__(self, parent: Optional[QObject], db_manager: Manager, wanted_classes: List[Any],
                 widget_creators: List[Callable], cls_black_list: Optional[List[type]] = None) -> None:
        """
        Construct the container.
        This specifies the database we are working on and the classes we can work with.

        Parameters
        ----------
        parent : Optional[QObject]
            The parent widget
        db_manager : Manager
            The Scine Database Manager
        wanted_classes : List[Any]
            The classes we want to work with, subclasses are automatically included
        widget_creators : List[Callable]
            The functions that can create widgets for the classes
        cls_black_list : Optional[List[type]], optional
            Specify classes that we don't want to be shown / selected, by default None

        Raises
        ------
        RuntimeError
            If the number of classes and the number of creators don't match
        """
        QWidget.__init__(self, parent)
        self.db_manager = db_manager
        self.class_searchers = [ChemotonClassSearcher(cls, black_list=cls_black_list) for cls in wanted_classes]
        self.doc_string_parser = DocStringParser()
        self._joiner: Optional[EngineJoiner] = None
        self._delete_join_threads: List[Thread] = []

        if len(wanted_classes) != len(widget_creators):
            raise RuntimeError("InternalError, invalid input")

        # take care of model and filters
        self._model_builder = ModelOptionsWidget()
        self._filter_builder = FilterBuilder()
        self._settings_builder = GeneralSettingsWidget()
        self._init_argument_builder_tabs = TabWidget(parent=self)
        self._init_argument_builder_tabs.addTab(self._model_builder, "Model")
        self._init_argument_builder_tabs.addTab(self._filter_builder, "Filter")
        self._init_argument_builder_tabs.addTab(self._settings_builder, "General Settings")
        self._init_argument_builder_tabs.setMaximumWidth(500)

        # generate the widgets to build widgets
        self.__add_chemoton_widgets = [creator(self, self.db_manager, searcher)
                                       for searcher, creator in zip(self.class_searchers, widget_creators)]

        # create widget for scroll area
        scroll_area_content = QWidget()
        self._scroll_area_content_layout = QVBoxLayout()
        # add toolbar
        self._toolbar = HeronToolBar()
        # play button
        self._play = self._toolbar.shortened_add_action(QStyle.SP_MediaPlay, "Start exploration",
                                                        "Ctrl+M", lambda: self._start_stop_all(True))
        self._stop = self._toolbar.shortened_add_action(QStyle.SP_MediaPause, "Stop exploration",
                                                        "Ctrl+R", lambda: self._start_stop_all(False))
        self._toolbar.shortened_add_action(
            QStyle.SP_DialogOpenButton, "Load Exploration Setup", "Ctrl+O", self._load_file
        )
        self._toolbar.shortened_add_action(
            QStyle.SP_DialogSaveButton, "Save Exploration Setup", "Ctrl+S", self._save_file
        )
        self._toggle_sound_action = self._toolbar.shortened_add_action(QStyle.SP_MediaVolumeMuted,
                                                                       "Mute/Unmute sounds",
                                                                       "Ctrl+P",
                                                                       self._toggle_sound)
        self.sound_allowed = False
        self._play.setCheckable(True)
        self._stop.setCheckable(True)
        self.is_blocked = False
        self._scroll_area_content_layout.addWidget(self._toolbar)
        # add holder of all the created widgets
        self.created_chemoton_widgets: List[StartStopWidget] = []
        self._created_chemoton_widgets_holder = QWidget()
        self.__created_chemoton_widgets_layout = QGridLayout()
        self._created_chemoton_widgets_holder.setLayout(self.__created_chemoton_widgets_layout)
        # finish scroll area content
        self._scroll_area_content_layout.addWidget(self._created_chemoton_widgets_holder)
        scroll_area_content.setLayout(self._scroll_area_content_layout)

        # create scroll area
        self._scroll_area = QScrollArea()
        self._scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self._scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self._scroll_area.setWidget(scroll_area_content)
        self._scroll_area.setWidgetResizable(True)

        # fill total grid layout of the tab
        self._grid_layout = QGridLayout()
        self._grid_layout.addWidget(self._init_argument_builder_tabs, 0, 0)
        self._grid_layout.addWidget(self._scroll_area, 0, 1)
        for i, widget in enumerate(self.__add_chemoton_widgets, 1):
            self._grid_layout.addWidget(widget, 1, i)

        self._max_widgets_per_row = 3
        self.setLayout(self._grid_layout)

    def closeEvent(self, event: QCloseEvent) -> None:
        """
        Clean up before we are closed
        """
        if self._joiner is not None:
            self._joiner.wait()
        for thread in self._delete_join_threads:
            thread.join()
        for widget in self.created_chemoton_widgets:
            widget.start_stop(stop_all=True)
        for widget in self.created_chemoton_widgets:
            widget.join(force_join=True)
        super().closeEvent(event)

    def _toggle_sound(self) -> None:
        self.sound_allowed = not self.sound_allowed
        self._toggle_sound_action.setIcon(
            self.style().standardIcon(QStyle.SP_MediaVolume)
            if self.sound_allowed
            else self.style().standardIcon(QStyle.SP_MediaVolumeMuted)
        )

    def _pre_io_operations(self) -> None:
        """
        Possible operations that we may want to do before we do disc I/O operations.
        Nothing currently implemented, but child classes can override this.
        """

    def _post_io_operations(self) -> None:
        """
        Possible operations that we may want to do after we do disc I/O operations.
        Nothing currently implemented, but child classes can override this.
        """

    def add_widget(self, chemoton_widget: StartStopWidget, predefined_position: bool = False) -> None:
        """
        Adds a new widget to the grid layout.

        Parameters
        ----------
        chemoton_widget : StartStopWidget
            The new widget
        predefined_position : bool, optional
            If the widget should be simply added or placed at a special position, by default False
            This addition is currently not implemented, but has to be done by child class.
            In this case, only the doc strings are set and the delete button is connected.
        """
        # connect functionalities
        chemoton_widget.set_docstring_dict(self.doc_string_parser)
        if chemoton_widget.button_delete is not None:
            chemoton_widget.button_delete.clicked.connect(  # pylint: disable=no-member
                partial(self.delete_widget, chemoton_widget)
            )
        if predefined_position:
            return
        # add widget
        count = len(self.created_chemoton_widgets)
        row = count // self._max_widgets_per_row
        column = count % self._max_widgets_per_row

        self.created_chemoton_widgets.append(chemoton_widget)
        self.__created_chemoton_widgets_layout.addWidget(chemoton_widget, row, column)

    def delete_widget(self, chemoton_widget: StartStopWidget) -> int:
        """
        Deletes a widget from the grid layout.

        Parameters
        ----------
        chemoton_widget : StartStopWidget
            The widget to be deleted

        Returns
        -------
        int
            The index of the deleted widget
        """
        idx = self.created_chemoton_widgets.index(chemoton_widget)
        self.created_chemoton_widgets.pop(idx)
        chemoton_widget.stop_class_if_working()
        self._delete_join_threads.append(Thread(target=chemoton_widget.join, kwargs={'force_join': True}))
        self._delete_join_threads[-1].start()

        chemoton_widget.setAttribute(Qt.WA_DeleteOnClose)
        chemoton_widget.close()
        chemoton_widget.setParent(None)  # type: ignore
        for widget in self.created_chemoton_widgets:
            self.__created_chemoton_widgets_layout.removeWidget(widget)

        # add remaining widgets
        n_widgets = len(self.created_chemoton_widgets)
        for i in range(n_widgets):
            row = i // self._max_widgets_per_row
            column = i % self._max_widgets_per_row
            self.__created_chemoton_widgets_layout.addWidget(self.created_chemoton_widgets[i], row, column)

        return idx

    def get_filters(self) -> Tuple[AggregateFilter, ReactiveSiteFilter, FurtherExplorationFilter, StructureFilter,
                                   ElementaryStepFilter, ReactionFilter, AggregateEnabling, AggregateFilter,
                                   ReactionEnabling, ReactionFilter, EnableCalculationResults, ReactionDisabling,
                                   StepDisabling]:
        """
        Returns the filters that are currently set.

        Returns
        -------
        Tuple[AggregateFilter, ReactiveSiteFilter, FurtherExplorationFilter, StructureFilter, ElementaryStepFilter,
                ReactionFilter, AggregateEnabling, AggregateFilter, ReactionEnabling, ReactionFilter,
                ReactionDisabling, StepDisabling]
            The filters as a tuple
        """
        return self._filter_builder.get_filters()

    def get_model(self) -> Model:
        """
        Return the current model.

        Returns
        -------
        Model
            The current model
        """
        return self._model_builder.model

    def get_settings(self) -> ValueCollection:
        """
        Returns the current settings.

        Returns
        -------
        ValueCollection
            The current settings
        """
        return ValueCollection(self._settings_builder.settings)

    def _start_stop_all(self, start: bool, force_join: bool = False) -> None:
        """
        Start or stop all widgets depending on the input.
        Stopping them also takes care of joining forked processes.

        Parameters
        ----------
        start : bool
            Whether we should start (True) or stop (False) all widgets
        force_join : bool, optional
            Propagates argument to engine, whether we should join also engines
            that have not been working, by default False
        """
        if not self.created_chemoton_widgets:
            write_error_message("You have not created any engines, yet")
            return
        if self.is_blocked:
            write_error_message("Currently in the process of stopping all engines, please be patient")
            return
        if start and force_join:
            write_error_message("Internal error, conflicting inputs")
            return
        button = self._play if start else self._stop
        button.setChecked(True)
        for widget in self.created_chemoton_widgets:
            if start:
                widget.start_stop(start_all=True)
            else:
                widget.start_stop(stop_all=True)
        # additional loop for shutdown to collect processes
        if not start:
            self.is_blocked = True
            self._joiner = EngineJoiner(self, force_join)
            self._joiner.start()
        button.setChecked(False)

    def _toggle_button(self, start: bool):
        """
        Toggle our play or stop button to signal interaction.

        Parameters
        ----------
        start : bool
            Whether we should toggle the play (True) or stop (False) button.
        """
        if start:
            self._play.toggle()
        else:
            self._stop.toggle()

    def _load_file(self) -> None:
        """
        Loads a widget setup from file.
        """
        filename = get_load_file_name(self, "exploration", ["json", "pickle", "pkl"])
        if filename is None:
            return
        self._pre_io_operations()
        if filename.suffix == ".json":
            with open(filename, "r") as f:
                data = f.read()
            save_object = json_wrap.decode(data)
        else:
            with open(filename, "rb") as f:
                save_object = pickle.load(f)
        self._unpack_save_object(save_object)
        self._post_io_operations()

    def _save_file(self, force_join: bool = True) -> None:
        """
        Write current widget setup to file.
        This requires to stop all widgets first, hence may take some time.
        """
        if not self.created_chemoton_widgets:
            write_error_message("You have not created anything, yet")
            return
        filename = get_save_file_name(self, "exploration", ["json", "pickle", "pkl"])
        if filename is None:
            return
        write_info_message("Stopping all engines, this may take some time...", timer=1_000_000_000)
        self._start_stop_all(start=False, force_join=force_join)
        # wait for all engines to be joined
        while self.is_blocked:
            sleep(0.1)
        clear_status_bar()
        self._pre_io_operations()
        write_info_message("Writing to disk")
        save_object = self._generate_save_object()
        try:
            if filename.suffix == ".json":
                encoded = json_wrap.encode(save_object)
                with open(filename, "w") as f:
                    f.write(encoded)
            else:
                with open(filename, "wb") as f:
                    pickle.dump(save_object, f)
        except BaseException as e:
            write_error_message(f"Error while saving: {e}")
        self._post_io_operations()

    def _generate_save_object(self) -> List[List[Any]]:
        """
        Puts all current widgets into one objects from which the widgets can be
        reconstructed.

        Returns
        -------
        List[List[Any]]
            The save object
        """
        save_object: List[List[Any]] = []
        for widget in self.created_chemoton_widgets:
            init_arguments = [type(widget)] + widget.init_arguments[:]
            save_object.append(init_arguments)
        save_object.append([self.get_model()])
        save_object = make_picklable(save_object)
        return save_object

    def _unpack_save_object(self, save_object: List[List[Any]]) -> None:
        """
        The inverse of _generate_save_object.

        Parameters
        ----------
        save_object : List[List[Any]]
            The object holding the reconstructed widget information.
        """
        model = save_object.pop(-1)[0]
        assert isinstance(model, Model)
        self._model_builder.model = model
        for init_arguments in save_object:
            cls = init_arguments.pop(0)
            for arg in init_arguments:
                if hasattr(arg, "initialize_collections"):
                    arg.initialize_collections(self.db_manager)
            inst = cls(*init_arguments)
            self.add_widget(inst)


class EngineJoiner(QThread):
    """
    A thread that joins all engines.
    """

    def __init__(self, parent: ChemotonWidgetContainer, force_join: bool = False):
        """
        Construct the thread.

        Parameters
        ----------
        parent : ChemotonWidgetContainer
            The container that holds all widgets that should be joined.
        force_join : bool, optional
            Whether engines should be joined also if they were not working, by default False
        """
        super().__init__(parent)
        self.container = parent
        self._force = force_join

    def run(self):
        """
        The thread's main loop, which joins all engines.
        """
        for widget in self.container.created_chemoton_widgets:
            widget.join(self._force)
        self.container.is_blocked = False
        self.exit(0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import Optional, Dict
from PySide2.QtWidgets import (
    QWidget,
    QVBoxLayout,
)
from PySide2.QtCore import QObject

from scine_heron.containers.combo_box import BaseBox


class ComboBoxTabWidget(QWidget):
    """
    Manually implements tabs and tab changes, by only displaying a combo box and change according to the box.
    We do this because a combo box requires less space than listing all tabs next to each other.
    """

    def __init__(self, tabs: Optional[Dict[str, Optional[QWidget]]] = None, parent: Optional[QObject] = None):
        """
        Construct this container widget with the underlying widgets.

        Parameters
        ----------
        tabs : Optional[Dict[str, QWidget]], optional
            The tabs we hold based on the text and the actual widget.
            Can also be None and be set later via setter, by default None.
        parent : Optional[QObject], optional
            The parent widget, by default None
        """
        super().__init__(parent=parent)
        self._box = BaseBox(self)
        self._layout = QVBoxLayout()
        self._layout.addWidget(self._box)
        self.setLayout(self._layout)
        if tabs is not None:
            self._tabs: Dict[str, Optional[QWidget]] = tabs
            self.set_tabs(tabs)
        else:
            self._tabs = {}

    def _change_tab(self) -> None:
        """
        Changes the displayed widget
        """
        for widget in self._tabs.values():
            if widget is not None:
                widget.setVisible(False)
        wanted = self._box.currentText()
        new_widget = self._tabs[wanted]
        if new_widget is not None:
            new_widget.setVisible(True)

    def set_tabs(self, tabs: Dict[str, Optional[QWidget]]) -> None:
        """
        Rebuilds the container widget with new subwidgets.

        Parameters
        ----------
        tabs : Dict[str, QWidget]
            The new subwidgets.
        """
        self._tabs = tabs
        try:
            self._box.currentIndexChanged.disconnect()  # pylint: disable=no-member
        except RuntimeError:
            pass
        self._box.clear()
        self._box.addItems(list(self._tabs.keys()))
        for key, widget in self._tabs.items():
            if widget is not None:
                self._layout.addWidget(widget)
                self._box.setCurrentText(key)
        self._box.setCurrentText(list(self._tabs.keys())[0])
        self._change_tab()
        self.updateGeometry()
        self._box.currentIndexChanged.connect(  # pylint: disable=no-member
            self._change_tab
        )

    def add_tab(self, tab_name: str, tab: Optional[QWidget]) -> None:
        """
        Adds a new tab to the widget.

        Parameters
        ----------
        tab_name : str
            The name of the tab to appear in the combo box.
        tab : QWidget
            The new tab.
        """
        self._tabs[tab_name] = tab
        self.set_tabs(self._tabs)

    def remove_tab(self, tab_name: str) -> None:
        """
        Removes a tab from the widget.

        Parameters
        ----------
        tab_name : str
            The tab to remove.
        """
        widget = self._tabs.pop(tab_name)
        if widget is not None:
            widget.deleteLater()
            widget.setParent(None)  # type: ignore
        self.set_tabs(self._tabs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the ControlledCheckBox class.
"""

import typing
from scine_heron.status_manager import WriteableStatus
from PySide2.QtWidgets import QCheckBox
from PySide2.QtCore import QObject, Qt


class ControlledCheckBox(QCheckBox):
    """
    A slightly modified version of Qt's QCheckBox that displays
    the state as defined by a model.
    """

    def __init__(
        self,
        status: WriteableStatus[Qt.CheckState],
        parent: typing.Optional[QObject] = None,
    ):
        """
        :param pull: Returns the current state of the model.
        :param push: Requests a change to the provided state.
        :param signal: Notifies when the state in model has changed.
        """
        super().__init__(parent)

        self.__configure_input_handling(status)
        self.__configure_output_handling(status)

        self.setCheckState(status.value)

    def __configure_output_handling(
        self, status: WriteableStatus[Qt.CheckState]
    ) -> None:
        """
        Configure how the view responds to changes in the model.
        """
        status.changed_signal.connect(self.setCheckState)

    def __configure_input_handling(
        self, status: WriteableStatus[Qt.CheckState]
    ) -> None:
        """
        Configure how the view responds to user input.
        """

        def handle_change(state: Qt.CheckState) -> None:
            """Handle user input."""
            blocked = self.signalsBlocked()
            self.blockSignals(True)
            try:
                self.setCheckState(status.value)
                status.value = state
            finally:
                self.blockSignals(blocked)

        self.stateChanged.connect(handle_change)  # pylint: disable=no-member



# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from os import path, walk
from itertools import combinations
from setuptools import setup, find_packages
import sys

min_version = (3, 6)
if sys.version_info < min_version:
    error = """
SCINE GUI does not support Python {0}.{1}.
Python {2}.{3} and above is required. Check your Python version like so:

python3 --version

This may be due to an out-of-date pip. Make sure you have pip >= 9.0.1.
Upgrade pip like so:

pip install --upgrade pip
""".format(
        *(sys.version_info[:2] + min_version)
    )
    sys.exit(error)

here = path.abspath(path.dirname(__file__))


def read_requirements(file_name: str):
    with open(path.join(here, file_name)) as requirements_file:
        # Parse requirements.txt, ignoring any commented-out lines.
        requirements = []
        for line in requirements_file:
            if line.startswith("#"):
                continue
            if line.startswith("-r"):
                requirements.extend(read_requirements(line.split()[-1].strip()))
            else:
                requirements.append(line.strip())
        return list(set(requirements))

variants = []
for subdir, dirs, files in walk(here):
    for f in files:
        if 'requirements-' in f:
            variants.append(f.split("requirements-")[-1].split(".txt")[0])

assert "all" in variants
assert "dev" in variants
variants.remove('all')
variants.remove('dev')
package_variants = {
    'all': read_requirements('requirements-all.txt')
}
variants_combinations = [["_".join(sorted(i)) for i in combinations(variants, len(variants) - j)]
                         for j in range(1, len(variants))]


def flatten_list(_list):
    return [item for sublist in _list for item in sublist]


flat_variants = flatten_list(variants_combinations)
for v in flat_variants:
    sub_variants = v.split("_")
    all_requirements = list(set(flatten_list(
        [read_requirements(f"requirements-{sub}.txt") for sub in sub_variants])))
    package_variants[v] = all_requirements

with open(path.join(here, "README.rst"), encoding="utf-8") as readme_file:
    readme = readme_file.read()

with open(path.join(here, 'scine_heron', '_version.py')) as f:
    exec(f.read())

# Define the setup
setup(
    name="scine_heron",
    version=__version__,
    author="ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group",
    author_email="scine@phys.chem.ethz.ch",
    description="Graphical user interface for SCINE",
    long_description=readme,
    url="https://www.scine.ethz.ch",
    python_requires=">={}".format(".".join(str(n) for n in min_version)),
    packages=find_packages(include=["scine_heron", "scine_heron.*"],
                           exclude=["scine_heron.tests*"]),
    include_package_data=True,
    package_data={
        "scine_heron": [
            "resources/*",
            "resources/icons/*",
            "resources/sounds/*",
            # When adding files here, remember to update MANIFEST.in as well,
            # or else they will not be included in the distribution on PyPI!
            # 'path/to/data_file',
        ]
    },
    license="BSD (3-clause)",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: C++",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Chemistry",
    ],
    entry_points={
        'console_scripts': [
            'scine_heron = scine_heron.__main__:main',
            'scine_heron_view_trajectory = scine_heron.__init__:view_trajectory_cli',
        ],
    },
    install_requires=read_requirements('requirements.txt'),
    extras_require=package_variants,
    zip_safe=False,
    test_suite="pytest",
    tests_require=read_requirements('requirements-all.txt'),
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from pathlib import Path
from os import path
from typing import Callable, Optional, List
from PySide2.QtWidgets import QFileDialog, QWidget

from scine_heron.utilities import write_error_message, write_info_message


def _valid_extension(file_name: Path, allowed_extensions: List[str]) -> bool:
    checked_extensions = [f'.{e}' if not e.startswith('.') else e for e in allowed_extensions]
    if file_name.suffix not in checked_extensions:
        write_error_message(f"Unsupported file extension '{file_name.suffix}', allowed are {checked_extensions}")
        return False
    return True


def get_load_file_name(parent: QWidget, default_name: str, valid_extensions: List[str],
                       validator: Optional[Callable[[Path], bool]] = None) -> Optional[Path]:
    """
    Ask the user to provide the name of the file to be loaded with a QFileDialog.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the QFileDialog
    default_name : str
        The default name of the file
    valid_extensions : List[str]
        Valid file extensions without leading '.'
    validator : Optional[Callable[[Path], bool]]
        An optional callable that carries out additional sanity checks on the file name
        and returns if it is valid (true) or not, by default None

    Notes
    -----
    * Not thread safe
    * A return value of None means that something was wrong, error or info has been written to the statusbar

    Returns
    -------
    Path
        The file name as a Path object
    """
    if not valid_extensions:
        valid_extensions = ["*"]
    filename, _ = QFileDialog.getOpenFileName(
        parent,
        parent.tr("Open File"),  # type: ignore[arg-type]
        default_name + "." + valid_extensions[0],
        parent.tr(f"{default_name} ({' *.'.join(['', *valid_extensions])[1:]})"),  # type: ignore[arg-type]
        # explanation last line: valid_extension = ["pkl", "json"] --> string after default_name = (*.pkl *json)
    )
    if not filename:
        write_info_message("Aborted loading")
        return None
    filename = Path(filename)
    if not path.exists(filename):
        write_error_message(f"File {filename} does not exist!")
        return None
    if not _valid_extension(filename, valid_extensions):
        return None
    if validator is not None and not validator(filename):
        return None
    return filename


def get_save_file_name(parent: QWidget, default_name: str, valid_extensions: List[str],
                       validator: Optional[Callable[[Path], bool]] = None) -> Optional[Path]:
    """
    Ask the user to provide the name of a file that should be written based on a QFileDialog.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the QFileDialog
    default_name : str
        The default name of the file
    valid_extensions : List[str]
        Valid file extensions without leading '.'
    validator : Optional[Callable[[Path], bool]]
        An optional callable that carries out additional sanity checks on the file name
        and returns if it is valid (true) or not, by default None

    Notes
    -----
    * Not thread safe
    * A return value of None means that something was wrong, error or info has been written to the statusbar

    Returns
    -------
    Path
        The file name as a Path object
    """
    if not valid_extensions:
        valid_extensions = ["*"]
    filename, _ = QFileDialog.getSaveFileName(
        parent,
        parent.tr("Save File"),  # type: ignore[arg-type]
        default_name + "." + valid_extensions[0],
        parent.tr(f"{default_name} ({' *.'.join(['', *valid_extensions])[1:]})"),  # type: ignore[arg-type]
    )
    if not filename:
        write_info_message("Aborted saving")
        return None
    filename = Path(filename)
    if not _valid_extension(filename, valid_extensions):
        return None
    if validator is not None and not validator(filename):
        return None
    return filename



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the DocStringParser class.
"""

import ast
import inspect
from itertools import tee
from typing import Dict, Tuple, Union, Any, Iterator

from numpydoc.docscrape import FunctionDoc


class DocStringParser:
    """
    Automatically parse Attribute Docstrings.
    """

    def __init__(self) -> None:
        self.__cache: Dict[str, Dict[str, str]] = {}

    def get_docstring_for_class_init(self, name: str, cls: object) -> Dict[str, str]:
        """
        Somewhat experimental function that tries to manually parse a doc string manually

        Parameters
        ----------
        name : str
            The name of the class
        cls : object
            The class object we want to parse the docstring for its __init__ method

        Returns
        -------
        Dict[str, str]
            The docstring with parameter name and description
        """
        name += "__init__"
        if name in self.__cache:
            return self.__cache[name]
        doc = cls.__init__.__doc__  # type: ignore
        if not doc:
            self.__cache[name] = {}
            return {}
        parsed_doc = FunctionDoc(cls.__init__)  # type: ignore
        params = parsed_doc["Parameters"]
        data = {}
        for p in params:
            desc = ""
            if p.type:
                desc += f"Type: {p.type}\n"
            if p.desc:
                desc += "\n".join(p.desc)
            data[p.name] = desc
        self.__cache[name] = data
        return data

    def get_docstring_for_instance_init(self, name: str, inst: object) -> Dict[str, str]:
        return self.get_docstring_for_class_init(name, inst.__class__)

    def get_docstring_for_object_attrs(self, name: str, o: object) -> Dict[str, str]:
        """
        Extracts PEP-224 docstrings for variables of `o`.
        """
        if name in self.__cache:
            return self.__cache[name]

        data: Dict[str, str] = dict()

        lines, _ = inspect.getsourcelines(o.__class__)
        doc = inspect.cleandoc("".join(["\n"] + lines))
        tree = ast.parse(doc).body[0]

        for node in reversed(tree.body):  # type: ignore[attr-defined]
            if isinstance(node, ast.FunctionDef) and node.name == "__init__":
                for assign_node, str_node in self.__pairwise(
                    ast.iter_child_nodes(node)
                ):
                    if not (
                        isinstance(assign_node, (ast.Assign, ast.AnnAssign))
                        and isinstance(str_node, ast.Expr)
                        and isinstance(str_node.value, ast.Str)
                    ):
                        continue
                    docstring = inspect.cleandoc(str_node.value.s).strip()
                    data[self.__get_name(assign_node)] = docstring

        self.__cache[name] = data

        return data

    @staticmethod
    def __pairwise(iterable: Iterator[Any]) -> Iterator[Tuple[Any, Any]]:
        """s -> (s0,s1), (s1,s2), (s2, s3), ..."""
        a, b = tee(iterable)
        next(b, None)
        return zip(a, b)

    @staticmethod
    def __get_name(assign: Union[ast.Assign, ast.AnnAssign]) -> Any:
        if isinstance(assign, ast.Assign) and len(assign.targets) == 1:
            target = assign.targets[0]
        elif isinstance(assign, ast.AnnAssign):
            target = assign.target
        else:
            return str()
        return target.attr  # type: ignore[attr-defined]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the energy profile class
"""

from datetime import datetime
from typing import Optional


class EnergyProfilePoint:
    def __init__(
        self, energy: float = 0, elapsed_time: float = 0, time_stamp: Optional[datetime] = None
    ):
        self.energy: float = energy
        self.elapsed_time: float = elapsed_time
        if time_stamp is None:
            self.time_stamp = datetime.now()
        else:
            self.time_stamp = time_stamp



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import Any


class DummyType(type):
    def __getattribute__(self, item):
        if item.startswith("__"):
            return super().__getattribute__(item)
        return super().__new__(DummyType, "DummyType", (type,), globals())


class DummyClass(metaclass=DummyType):

    def __init__(self, *args, **kwargs):
        self.cls = None

    def __getattribute__(self, item):
        if item == "__class__":
            return object.__getattribute__(self, "__class__")
        try:
            return object.__getattribute__(self, "cls")
        except AttributeError:
            return object.__getattribute__(self, "__class__")

    def __setattr__(self, key, value):
        pass

    def __getitem__(self, item):
        return getattr(self, "cls")

    def __setitem__(self, key, value):
        pass

    def __call__(self, *args, **kwargs):
        pass

    def __repr__(self):
        return f"{self.__class__.__name__}()"


class NotMatchingClass(DummyClass):
    """
    This class is returned as a placeholder for any demanded module/class
    that is an optional dependency and is not installed.
    Its main task is to not match any types in type checks.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cls = self.__class__


class AnyMatchingClass(DummyClass):
    """
    Same idea as NotMatchClass, but with the idea to match every type check.
    However, any returned variables did work as typehints for mypy, so currently unused.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cls = Any



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import Optional
from copy import deepcopy
from PySide2.QtCore import QObject
from PySide2.QtWidgets import (
    QDialog,
    QLabel
)

from scine_chemoton.utilities.model_combinations import ModelCombination

from scine_heron.containers.layouts import VerticalLayout
from scine_heron.settings.class_options_widget import ModelOptionsWidget


class ModelCombinationBuilder(QDialog):
    def __init__(self, options: ModelCombination, parent: Optional[QObject] = None):
        """
        Define a model combination.
        """
        super().__init__(parent=parent)
        # widgets
        self._electronic_label = QLabel("Model for electronic energies")
        self._electronic_model_widget = ModelOptionsWidget(parent=parent, model=options.electronic_model)
        self._hessian_label = QLabel("Model for free energy corrections")
        self._hessian_model_widget = ModelOptionsWidget(parent=parent, model=options.hessian_model)
        # layout
        layout = VerticalLayout([
            self._electronic_label,
            self._electronic_model_widget,
            self._hessian_label,
            self._hessian_model_widget])
        self.setLayout(layout)

    def get_model_combination(self) -> ModelCombination:
        model_combination = ModelCombination(
            deepcopy(self._electronic_model_widget.model),
            deepcopy(self._hessian_model_widget.model)
        )
        return model_combination



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

"""
Provides the DictOptionWidget class.
"""
import inspect
import pickle
import yaml
from collections import UserDict, UserList
from copy import deepcopy
from enum import Enum
from itertools import product
from os import path
from typing import Optional, Dict, Any, Union, List, Tuple, Set, Type, Callable, Iterable, TYPE_CHECKING

from scine_utilities import ValueCollection, Settings, OptionListDescriptor

from PySide2.QtCore import Qt
from PySide2.QtGui import QCloseEvent
from PySide2.QtWidgets import (
    QVBoxLayout,
    QDialog,
    QWidget,
    QFrame,
    QLabel,
    QSpinBox,
    QLineEdit,
    QDoubleSpinBox,
    QCheckBox,
    QGridLayout,
    QPushButton,
    QStyle,
    QCompleter,
    QHBoxLayout
)

from scine_heron.containers.wrapped_label import WrappedLabel
from scine_heron.containers.buttons import TextPushButton
from scine_heron.containers.layouts import VerticalLayout
from scine_heron.containers.without_wheel_event import NoWheelSpinBox, NoWheelDoubleSpinBox, NoWheelComboBox
from scine_heron.io.file_browser_popup import get_save_file_name, get_load_file_name
from scine_heron.io.json_pickle_wrap import encode, decode
from scine_heron.io.text_box import text_input_box
from scine_heron.settings.enum_selection import EnumSelectionBox
from scine_heron.toolbar.io_toolbar import ToolBarWithSaveLoad
from scine_heron.utilities import write_error_message, write_info_message
from scine_heron.dependencies.optional_import import importer, is_imported
from scine_heron.settings.docstring_parser import DocStringParser

if TYPE_CHECKING:
    from scine_database import Model, Job, Side, Label, ID
    from scine_chemoton.utilities.options import BaseOptions
    from scine_chemoton.reaction_rules import RuleSet, BaseRule
    from scine_chemoton.steering_wheel.datastructures import GearOptions, LogicCoupling
    from scine_heron.chemoton.rule_builder import RuleBuilderButtonWrapper, RuleSetBuilderButtonWrapper
    from scine_chemoton.utilities.model_combinations import ModelCombination
    from scine_chemoton.utilities.db_object_wrappers.thermodynamic_properties import ReferenceState
    from scine_heron.chemoton.reference_state_builder import ReferenceStateBuilder
    from scine_heron.chemoton.gear_options_widget import GearOptionsBuilderButtonWrapper
    from scine_heron.chemoton.side_builder import SideBuilder
    from scine_chemoton.gears.kinetic_modeling.atomization import (
        MultiModelEnergyReferences, PlaceHolderMultiModelEnergyReferences
    )
    from scine_chemoton.utilities.uncertainties import UncertaintyEstimator, ZeroUncertainty
    from scine_chemoton.utilities.place_holder_model import construct_place_holder_model
    from scine_heron.chemoton.energy_reference_builder import EnergyReferenceBuilder
    from scine_heron.chemoton.uncertainty_tuple import UncertaintyTuple
    from scine_chemoton.utilities.reactive_complexes.inter_reactive_complexes import InterReactiveComplexes
else:
    Model = importer("scine_database", "Model")
    Side = importer("scine_database", "Side")
    Job = importer("scine_database", "Job")
    Label = importer("scine_database", "Label")
    ID = importer("scine_database", "ID")
    BaseOptions = importer("scine_chemoton.utilities.options", "BaseOptions")
    ModelCombination = importer("scine_chemoton.utilities.model_combinations", "ModelCombination")
    RuleSet = importer("scine_chemoton.reaction_rules", "RuleSet")
    BaseRule = importer("scine_chemoton.reaction_rules", "BaseRule")
    RuleBuilderButtonWrapper = importer("scine_heron.chemoton.rule_builder", "RuleBuilderButtonWrapper")
    RuleSetBuilderButtonWrapper = importer("scine_heron.chemoton.rule_builder", "RuleSetBuilderButtonWrapper")
    ReferenceState = importer("scine_chemoton.utilities.db_object_wrappers.thermodynamic_properties", "ReferenceState")
    MultiModelEnergyReferences = importer("scine_chemoton.gears.kinetic_modeling.atomization",
                                          "MultiModelEnergyReferences")
    PlaceHolderMultiModelEnergyReferences = importer("scine_chemoton.gears.kinetic_modeling.atomization",
                                                     "PlaceHolderMultiModelEnergyReferences")
    ZeroUncertainty = importer("scine_chemoton.utilities.uncertainties", "ZeroUncertainty")
    construct_place_holder_model = importer("scine_chemoton.utilities.place_holder_model",
                                            "construct_place_holder_model")
    UncertaintyEstimator = importer("scine_chemoton.utilities.uncertainties", "UncertaintyEstimator")
    ReferenceStateBuilder = importer("scine_heron.chemoton.reference_state_builder", "ReferenceStateBuilder")
    SideBuilder = importer("scine_heron.chemoton.side_builder", "SideBuilder")
    GearOptions = importer("scine_chemoton.steering_wheel.datastructures", "GearOptions")
    LogicCoupling = importer("scine_chemoton.steering_wheel.datastructures",
                             "LogicCoupling")  # pylint: disable=unused-wildcard-import
    GearOptionsBuilderButtonWrapper = importer("scine_heron.chemoton.gear_options_widget",
                                               "GearOptionsBuilderButtonWrapper")
    EnergyReferenceBuilder = importer("scine_heron.chemoton.energy_reference_builder", "EnergyReferenceBuilder")
    UncertaintyTuple = importer("scine_heron.chemoton.uncertainty_tuple", "UncertaintyTuple")
    InterReactiveComplexes = importer("scine_chemoton.utilities.reactive_complexes.inter_reactive_complexes",
                                      "InterReactiveComplexes")


class DefaultMapping:
    """
    Defines defaults for each type
    """

    def __init__(self) -> None:
        self._mapping: Dict[Type, Any] = {
            int: 0,
            bool: False,
            float: 0.0,
            str: "",
            dict: {},
            Dict: {},
            list: [],
            List: [],
            set: set(),
            Set: set(),
            tuple: tuple(),
            Tuple: tuple(),  # type: ignore
            UserDict: {},
            RuleSet: RuleSet({}),
            GearOptions: GearOptions(),
            UserList: [],
            ValueCollection: ValueCollection({}),
            Model: Model("PM6", "PM6", ""),
            Job: Job("job_order"),
            Label: Label.MINIMUM_OPTIMIZED,
            ID: ID(),
            Any: "",  # type: ignore
            type(None): None,
            Side: Side.BOTH,
        }
        if is_imported(Model):
            self._mapping[ReferenceState] = ReferenceState(float(Model("PM6", "PM6", "").temperature),
                                                           float(Model("PM6", "PM6", "").pressure))
            self._mapping[ModelCombination] = ModelCombination(construct_place_holder_model())
            self._mapping[MultiModelEnergyReferences] = PlaceHolderMultiModelEnergyReferences()
            self._mapping[UncertaintyEstimator] = ZeroUncertainty()
        if is_imported(UncertaintyTuple):
            self._mapping[UncertaintyTuple] = UncertaintyTuple.get_default()
        standard_types = self.standard_types()
        for t in standard_types:
            self._mapping[List[t]] = []  # type: ignore
        self._mapping[List[Any]] = []  # type: ignore
        for s in standard_types:
            for t in standard_types:
                self._mapping[Dict[s, t]] = {}  # type: ignore
                self._mapping[Dict[s, Any]] = {}  # type: ignore
                self._mapping[Dict[Any, t]] = {}  # type: ignore
        self._mapping[Dict[Any, Any]] = {}  # type: ignore

    def __getitem__(self, item: Optional[Type]):
        # TODO maybe cycle through keys and check for subclasses if class not explicitly present in map
        if item is None:
            return deepcopy(self._mapping[type(item)])
        return deepcopy(self._mapping[item])  # deepcopy ensures safety of mutable default types

    @staticmethod
    def standard_types() -> List[Type]:
        return [bool, int, float, str]

    @staticmethod
    def enhanced_standard_types() -> List[Type]:
        return DefaultMapping.standard_types() + [dict, list, set, tuple, ValueCollection, Model, Job, RuleSet,
                                                  Dict[Any, Any], List[Any], Set[Any], Tuple[Any, ...]]  # type: ignore

    def keys(self):
        return self._mapping.keys()


class DictOptionWidget(QFrame):
    """
    DictOptionWidget create a container of widgets based on options types.
    """

    def __init__(
            self,
            options: Dict[str, Any],
            parent: Optional[QWidget],
            show_border: Optional[bool] = False,
            docstring_dict: Optional[Dict[str, str]] = None,
            add_close_button: bool = True,
            allow_additions: bool = False,
            allow_removal: bool = True,
            value_type: Optional[Type] = None,
            addition_suggestions: Optional[List[str]] = None,
            suggestions_by_name: Optional[Dict[str, Dict[str, Callable]]] = None,
            suggestions_with_values: Optional[Dict[str, Callable]] = None,
            keys_excluded_from_io: Optional[List[str]] = None,
    ) -> None:
        super(DictOptionWidget, self).__init__(parent)
        self._ignore_list = ["self", "args", "kwargs", "_"]
        self.__io_extensions = ['yaml', 'json', 'pkl', 'pickle']
        self._options = options
        self._option_widgets: Dict[str, QWidget] = {}
        self._option_getters: Dict[str, Callable[[], Any]] = {}
        self._docstring_dict = docstring_dict
        self._value_type = value_type
        self._allow_removal = allow_removal
        self._addition_suggestions = addition_suggestions
        self._suggestions_by_name = suggestions_by_name
        self._suggestions_with_values = suggestions_with_values
        self._keys_excluded_from_io = keys_excluded_from_io

        self.__default_min_value = -1000000000
        self.__default_max_value = 1000000000
        self.__double_spin_step = 0.1
        self.__double_spin_decimals = 10

        self.__standard_types = DefaultMapping.standard_types()
        self.__enhanced_standard_types = DefaultMapping.enhanced_standard_types()
        self.__default_mapping = DefaultMapping()

        if show_border:
            self.setFrameStyle(QFrame.StyledPanel | QFrame.Plain)

        self.__super_layout = QGridLayout()
        if allow_removal:
            self.__super_layout.addWidget(ToolBarWithSaveLoad(self._save, self._load, self))
        self.__layout = QGridLayout()
        for option_name, option in self._options.items():
            self.add_key_value(option_name, option)
        self.__options_holder_widget = QWidget()
        self.__options_holder_widget.setLayout(self.__layout)
        self.__super_layout.addWidget(self.__options_holder_widget)

        if allow_additions:
            self.__super_layout.addWidget(TextPushButton("Add new field", self.__add_option, self))
        if parent is not None and add_close_button:
            self.__super_layout.addWidget(
                TextPushButton("Ok", parent.close, self, shortcut="Return")  # type: ignore
            )

        self.setLayout(self.__super_layout)

    def add_key_value(self, option_name: str, option: Any, allow_removal_for_dict: bool = True) -> None:
        if option_name in self._ignore_list:
            return
        self._options[option_name] = option
        if isinstance(self._options, Settings) \
                and isinstance(self._options.descriptor_collection[option_name], OptionListDescriptor):
            assert option in self._options.descriptor_collection[option_name].options
            widget = self.__generate_selection_box(self._options.descriptor_collection[option_name].options)
            widget.setCurrentText(option)

            def getter():
                return widget.currentText()
        else:
            widget, getter = self.construct_widget_based_on_type(option, option_name, allow_removal_for_dict)
        self._option_widgets[option_name] = widget
        self._option_getters[option_name] = getter

        name_widget = WrappedLabel(option_name)
        if self._docstring_dict and option_name in self._docstring_dict:
            name_widget.setToolTip(self._docstring_dict[option_name])

        index = len(self._option_widgets)
        self.__layout.addWidget(name_widget, index, 0)
        self.__layout.addWidget(widget, index, 1)
        if self._allow_removal:
            self.__layout.addWidget(self.__generate_removal_widget(index), index, 2)

    def construct_widget_based_on_type(self, option: Any, option_name: Optional[str] = None,
                                       allow_removal_for_dict: bool = True) \
            -> Tuple[QWidget, Callable[[], Any]]:
        # base types
        # bool must be checked before int!
        if self.__option_check(option, [bool])[0]:
            option = self.__sanitize_option(option)
            widget = self.__generate_checkbox_widget(option)
            return widget, widget.isChecked
        if self.__option_check(option, [int])[0]:
            option = self.__sanitize_option(option)
            widget = self.__generate_spin_widget(option)
            return widget, widget.value
        if self.__option_check(option, [float])[0]:
            option = self.__sanitize_option(option)
            widget = self.__generate_double_spin_widget(option)
            return widget, widget.value
        if self.__option_check(option, [str])[0]:
            option = self.__sanitize_option(option)
            widget = self.__generate_line_editor_widget(option)
            if self._suggestions_by_name is not None and option_name is not None \
                    and option_name in self._suggestions_by_name:
                widget.setCompleter(QCompleter(list(self._suggestions_by_name[option_name].keys())))
            return widget, widget.text

        # Scine types
        if self.__option_check(option, [ID])[0]:
            option = self.__sanitize_option(option)
            widget = self.__generate_line_editor_widget(str(option))

            def id_getter() -> ID:
                return ID(widget.text())

            return widget, id_getter
        if self.__option_check(option, [RuleSet])[0]:
            option = self.__sanitize_option(option)
            widget = RuleSetBuilderButtonWrapper(option, self)
            return widget, widget.get_options
        if self.__option_check(option, [BaseRule])[0]:
            option_type = self.__option_check(option, [BaseRule])[1]
            if option_type is not None:
                widget = RuleBuilderButtonWrapper(option_type, self)
                return widget, widget.get_rule
        if self.__option_check(option, [Enum])[0]:
            option_type = self.__option_check(option, [Enum])[1]
            if option_type is not None and issubclass(option_type, Enum):
                widget = EnumSelectionBox(self, option_type)
                return widget, widget.get_value
        if self.__option_check(option, [Label])[0]:
            option_type = self.__option_check(option, [Label])[1]
            if option_type is not None and issubclass(option_type, Label):
                widget = EnumSelectionBox(self, option_type)  # type: ignore
                return widget, widget.get_value
        if self.__option_check(option, [GearOptions])[0]:
            option = self.__sanitize_option(option)
            gears = None
            if self._suggestions_by_name is not None and option_name is not None:
                suggestions = self._suggestions_by_name.get(option_name)
                if suggestions is not None and len(suggestions) == 1:
                    # only one option, we rely on implicit coupling that this single function gives us all gears
                    # TODO maybe make this more explicit
                    gears = list(suggestions.values())[0]()
            if not is_imported(GearOptionsBuilderButtonWrapper):
                raise ImportError("GearOptionsBuilderButtonWrapper could not be imported.")
            widget = GearOptionsBuilderButtonWrapper(option, gears, self)
            return widget, widget.get_options
        if self.__option_check(option, [ValueCollection])[0]:
            option = self.__sanitize_option(option)
            if option is None:
                option = self.__default_mapping[ValueCollection]
            if self._suggestions_by_name is not None and option_name is not None:
                suggest_dict = self._suggestions_by_name.get(option_name)
            else:
                suggest_dict = None
            widget = DictOptionWidget(
                option.as_dict(), parent=self, show_border=True, add_close_button=False, allow_additions=True,
                suggestions_with_values=suggest_dict, allow_removal=allow_removal_for_dict
            )

            def vc_getter():
                return ValueCollection(widget.get_widget_data())

            return widget, vc_getter
        if self.__option_check(option, [ModelCombination])[0]:
            # This class uses something from the dict_options_widget. Therefore, we cannot
            # import it on the top of the file because this would lead to circular imports.
            from scine_heron.chemoton.model_combination_builder import ModelCombinationBuilder
            option = self.__sanitize_option(option)
            widget = ModelCombinationBuilder(option, self)
            return widget, widget.get_model_combination

        if self.__option_check(option, [MultiModelEnergyReferences])[0]:
            option = self.__sanitize_option(option)
            widget = EnergyReferenceBuilder(option, self)
            return widget, widget.get_energy_reference

        if self.__option_check(option, [UncertaintyTuple])[0]:
            # This class uses something from the dict_options_widget. Therefore, we cannot
            # import it on the top of the file because this would lead to circular imports.
            from scine_heron.chemoton.uncertainty_estimator_builder import UncertaintyBuilder
            option = self.__sanitize_option(option)
            widget = UncertaintyBuilder(option, self)
            return widget, widget.get_uncertainty_tuple
        if self.__option_check(option, [UncertaintyEstimator])[0]:
            # This class uses something from the dict_options_widget. Therefore, we cannot
            # import it on the top of the file because this would lead to circular imports.
            from scine_heron.chemoton.uncertainty_estimator_builder import UncertaintyEstimatorBuilder
            option = self.__sanitize_option(option)
            widget = UncertaintyEstimatorBuilder(option, self)
            return widget, widget.get_uncertainty_estimator

        if self.__option_check(option, [Side])[0]:
            option = self.__sanitize_option(option)
            widget = SideBuilder(self)
            return widget, widget.get_side

        if self.__option_check(option, [ReferenceState])[0]:
            option = self.__sanitize_option(option)
            widget = ReferenceStateBuilder(option, self)
            return widget, widget.get_reference_state

        if self.__option_check(option, [Model, Job])[0]:
            san_option = self.__sanitize_option(option)
            if san_option is None:
                san_option = self.__default_mapping[self.__option_check(option, [Model, Job])[1]]
            widget = DictOptionWidget(
                self.get_attributes_of_object(san_option), parent=self, show_border=True, add_close_button=False,
                allow_removal=False
            )

            def mj_getter():
                self.set_attributes_to_object(san_option, widget.get_widget_data())
                return san_option

            return widget, mj_getter
        if self.__option_check(option, [BaseOptions])[0]:
            widget = DictOptionWidget(
                self.get_attributes_of_object(option), parent=self, show_border=True, add_close_button=False,
                allow_removal=False
            )

            def option_getter():
                self.set_attributes_to_object(option, widget.get_widget_data())
                return option

            return widget, option_getter

        if self.__option_check(option, [InterReactiveComplexes])[0]:
            san_option = self.__sanitize_option(option)
            if hasattr(san_option, "options"):
                widget_options = san_option.options
                obj = san_option
            elif hasattr(option, "options"):
                widget_options = option.options
                obj = option
            else:
                write_error_message("Received InterReactiveComplexes but could not find options")
                return QLabel("Error"), lambda: None
            parser = DocStringParser()
            doc_name = option_name if option_name is not None else InterReactiveComplexes.__name__
            doc_strings = parser.get_docstring_for_object_attrs(doc_name, obj)
            widget = DictOptionWidget(self.get_attributes_of_object(widget_options),
                                      parent=self, docstring_dict=doc_strings, add_close_button=False,
                                      show_border=True, allow_additions=False, allow_removal=False)

            def irc_getter():
                self.set_attributes_to_object(widget_options, widget.get_widget_data())
                obj.options = widget_options
                return obj

            return widget, irc_getter

        # iterable types
        if self.__option_check(option, [tuple, Tuple])[0]:  # type: ignore
            san_option = self.__sanitize_option(option)
            option_type = self.__first_non_union_type(option)
            widget = self.__generate_iterable_edit_widget(san_option, option_type, allow_additions=False)
            return widget, widget.get_values
        if self.__option_check(option, [list, List, UserList])[0]:
            san_option = self.__sanitize_option(option)
            # in the case of the annotation
            # Union[List[SomeType], None] we want to get the SomeType
            # in the case of the annotation
            # List[SomeType] we want to get the SomeType
            if isinstance(option, inspect.Parameter) and option.annotation != inspect.Parameter.empty \
                    and getattr(option.annotation, "__origin__", None) is Union:
                # we got a Union
                option_type = self.__first_non_union_type(self.__first_non_union_type(option))
            else:
                # we got a List directly
                option_type = self.__first_non_union_type(option)
            widget = self.__generate_iterable_edit_widget(san_option, option_type, allow_additions=True)
            return widget, widget.get_values
        if self.__option_check(option, [dict, Dict, UserDict])[0]:
            option_type = self.__option_check(option, [dict, Dict, UserDict])[1]
            if hasattr(option_type, "__args__") and len(getattr(option_type, "__args__")) == 2:
                value_type = getattr(option_type, "__args__")[1]
            else:
                value_type = None
            option = self.__sanitize_option(option)
            if option is None:
                option = self.__default_mapping[dict]
            if self._suggestions_by_name is not None and option_name is not None:
                suggestions = self._suggestions_by_name.get(option_name)
            else:
                suggestions = None
            widget = DictOptionWidget(option, parent=self, show_border=True, add_close_button=False,
                                      allow_additions=True, value_type=value_type, suggestions_with_values=suggestions,
                                      allow_removal=allow_removal_for_dict
                                      )
            return widget, widget.get_widget_data
        if self.__option_check(option, [Iterable])[0]:
            option_type = self.__first_non_union_type(option)
            widget = self.__generate_iterable_edit_widget([], option_type, allow_additions=True)
            return widget, widget.get_values

        # assume we have a non-default class that provides default constructor with options
        try:
            print(f"Warning: Could not deduce type of {option}, the widget might be incorrect.")
            attr = self.get_attributes_of_object(option)
            if hasattr(option, "options"):
                attr = {**attr, **self.get_attributes_of_object(option.options)}
            if self._suggestions_by_name is not None and option_name is not None:
                suggestions = self._suggestions_by_name.get(option_name)
            else:
                suggestions = None
            widget = DictOptionWidget(
                attr, parent=self, show_border=True, add_close_button=False, allow_removal=False,
                suggestions_with_values=suggestions
            )
            return widget, widget.get_widget_data
        except BaseException as e:  # TODO maybe remove exception info for release
            option_info = option if option is type else type(option)
            widget = QLabel(f"Type '{str(option_info)}' is not implemented: {str(e)}")
            return widget, lambda: None

    def __option_check(self, parameter: Any, types_to_check: List[Type]) \
            -> Tuple[bool, Optional[Type]]:
        """
        Also modifies parameter in case it is an inspect.Parameter to obtain a real type
        """
        for type_to_check in types_to_check:
            if isinstance(parameter, type_to_check):
                return True, type(parameter)
            # if received directly a type:
            if parameter == type_to_check:
                return True, parameter
        # add subscriptions to typing classes
        if List in types_to_check:
            types_to_check += [List[t] for t in self.__enhanced_standard_types]  # type: ignore
            types_to_check.append(List[Any])
        if Dict in types_to_check:
            for s in self.__standard_types:
                for t in self.__standard_types:
                    types_to_check.append(Dict[s, t])  # type: ignore
                    types_to_check.append(Dict[s, Any])  # type: ignore
                    types_to_check.append(Dict[Any, t])  # type: ignore
            types_to_check.append(Dict[Any, Any])  # type: ignore
        if Tuple in types_to_check and hasattr(parameter, "__args__"):
            n_args = len(getattr(parameter, "__args__"))
            for perm in product(list(self.__default_mapping.keys()), repeat=n_args):
                # for perm in product(list(self.__default_mapping.keys()), repeat=2):
                types_to_check.append(Tuple[perm])  # type: ignore
        # add Optional to all types
        types_to_check += [Optional[t] for t in types_to_check]  # type: ignore

        def types_equal(a: type, b: type) -> bool:
            if b == Iterable:
                a_origin = getattr(a, "__origin__", None)
                if a_origin is not None:
                    return issubclass(a_origin, Iterable)
            try:
                if issubclass(a, b):
                    return True
            except TypeError:
                pass
            return a == b

        for type_to_check in types_to_check:
            if parameter == type_to_check:
                return True, parameter
            if hasattr(parameter, "__origin__") and getattr(parameter, "__origin__") is Union:
                for arg in getattr(parameter, "__args__"):
                    if types_equal(arg, type_to_check):
                        return True, type_to_check
        # now it has to be an inspect.Parameter to still work
        if not isinstance(parameter, inspect.Parameter):
            return False, None

        def from_default(_parameter: inspect.Parameter, _types_to_check):
            # try to get the type from the assigned default
            default = _parameter.default
            if default == inspect.Parameter.empty:
                return False, None
            for _type_to_check in _types_to_check:
                if not hasattr(_type_to_check, "__origin__"):
                    # typing classes cannot be used in 'isinstance'
                    if isinstance(default, _type_to_check):
                        return True, type(default)
            return False, None

        annot = parameter.annotation
        # if there is a "from __future__ import annotations" call somewhere in our import tree
        # then the inspect.Parameter.annotation will not return the actual type, e.g., db.Model,
        # but the string of the type, e.g., "db.Model"
        # so, if we get a string, we try to reverse this with an eval call
        if isinstance(annot, str):
            try:
                annot = eval(annot)  # pylint: disable=eval-used
            except NameError as e:
                print(f"Error in evaluation of type annotation: {str(e)}")
                # cannot use annotation, try default
                return from_default(parameter, types_to_check)

        if Tuple in types_to_check and hasattr(annot, "__args__"):
            n_args = len(getattr(annot, "__args__"))
            for perm in product(list(self.__default_mapping.keys()), repeat=n_args):
                # for perm in product(list(self.__default_mapping.keys()), repeat=2):
                types_to_check.append(Tuple[perm])  # type: ignore

        if annot != inspect.Parameter.empty:

            # check if we have a Union
            if hasattr(annot, "__origin__") and getattr(annot, "__origin__") is Union:
                args = getattr(annot, "__args__")
                for type_to_check in types_to_check:
                    for arg in args:
                        if types_equal(arg, type_to_check):
                            return True, arg
            else:
                for type_to_check in types_to_check:
                    if types_equal(annot, type_to_check):
                        return True, annot
        # maybe no annotation but a default
        return from_default(parameter, types_to_check)

    def __sanitize_option(self, option: Any) -> Any:

        # if there is a "from __future__ import annotations" call somewhere in our import tree
        # then the inspect.Parameter.annotation will not return the actual type, e.g., db.Model,
        # but the string of the type, e.g., "db.Model"
        # so, if we get a string, we try to reverse this with an eval call
        def destringify_annot(annotation: Union[str, Type]) -> Type:
            if isinstance(annotation, str):
                try:
                    return eval(annotation)  # pylint: disable=eval-used
                except NameError as e:
                    # cannot use annotation, try default
                    raise RuntimeError("Cannot transform string annotation to type") from e
            return annotation

        if isinstance(option, inspect.Parameter):
            default = option.default
            if default != inspect.Parameter.empty:
                if default is None:
                    annot = destringify_annot(option.annotation)
                    if annot != inspect.Parameter.empty:
                        if hasattr(annot, "__origin__") and getattr(annot, "__origin__") is Union:
                            args = getattr(annot, "__args__")
                            for arg in args:
                                try:
                                    return self.__default_mapping[arg]
                                except KeyError:
                                    pass
                        return self.__default_mapping[annot]
                return default
            annot = destringify_annot(option.annotation)
        else:
            annot = option
        # check if we have a Union
        if hasattr(annot, "__origin__") and getattr(annot, "__origin__") is Union:
            args = getattr(annot, "__args__")
            # special case for rule building
            for arg in args:
                if hasattr(arg, "__origin__") or arg == Any:
                    continue
                if issubclass(arg, RuleSet) or issubclass(arg, BaseRule):
                    return arg({})
            # default case
            for arg in args:
                # roundabout way, because we only overwrite __getitem__ in DefaultMapping
                try:
                    return self.__default_mapping[arg]
                except KeyError:
                    pass
        # check if we have a Tuple
        if hasattr(annot, "__origin__") and getattr(annot, "__origin__") is tuple:
            args = getattr(annot, "__args__")
            return tuple(self.__default_mapping[arg] for arg in args)
        try:
            return self.__default_mapping[annot]
        except (KeyError, TypeError):
            return annot

    @staticmethod
    def __first_non_union_type(option: Any) -> type:
        if isinstance(option, inspect.Parameter):
            val = option.annotation
            if val == inspect.Parameter.empty:
                val = option.default
                if val == inspect.Parameter.empty:
                    return Any  # type: ignore
        else:
            val = option
        args = getattr(val, "__args__", None)
        if args is None:
            if isinstance(val, type):
                return val
            # check if we got a value instead of a type annotation and can get the type from its member
            if len(val):
                return type(val[0])
            return Any  # type: ignore
        # make sure we don't have a union
        while getattr(args[0], "__origin__", None) is Union:
            args = getattr(args[0], "__args__")
        return args[0]

    def __add_option(self) -> None:
        if self._value_type == Any:
            self._value_type = None
        require_value_type = self._value_type is None and self._suggestions_with_values is None
        if self._addition_suggestions is not None:
            suggestions = self._addition_suggestions
        elif self._suggestions_with_values is not None:
            suggestions = list(self._suggestions_with_values.keys())
        else:
            suggestions = None
        edit = OptionsAdditionDialog(add_value_field=require_value_type, parent=self, suggestions=suggestions)
        edit.exec_()
        name = edit.name.strip()
        if name:
            if self._suggestions_with_values is not None:
                value = self._suggestions_with_values.get(name)
                if value is None:
                    write_error_message(f"Entered invalid key {name}")
                else:
                    try:
                        self.add_key_value(name, value(), allow_removal_for_dict=False)
                    except (NameError, TypeError):
                        write_error_message(f"Could not construct {value} for key {name}")
            elif self._value_type is None:
                if edit.type.strip():
                    try:
                        cls = eval(edit.type.strip())  # pylint: disable=eval-used
                        self.add_key_value(name, cls())
                    except NameError:
                        write_error_message(f"Given type {edit.type} is not a valid type")
                    except TypeError:
                        write_info_message(f"Trying to infer type from {edit.type}")
                        self.add_key_value(name, cls)
                else:
                    self.add_key_value(name, "")
            else:
                self.add_key_value(name, self._value_type())  # type: ignore
        else:
            write_error_message("Did not enter an option name")

    def __generate_spin_widget(self, option: int) -> QSpinBox:
        spin_edit = NoWheelSpinBox()
        spin_edit.setMinimum(int(self.__default_min_value))
        spin_edit.setMaximum(int(self.__default_max_value))
        spin_edit.setValue(option)
        return spin_edit

    def __generate_double_spin_widget(self, option: float) -> QDoubleSpinBox:
        spin_edit = NoWheelDoubleSpinBox()
        spin_edit.setMinimum(float(self.__default_min_value))
        spin_edit.setMaximum(float(self.__default_max_value))
        spin_edit.setSingleStep(self.__double_spin_step)
        spin_edit.setDecimals(self.__double_spin_decimals)
        spin_edit.setValue(option)
        return spin_edit

    def __generate_iterable_edit_widget(self, option: Iterable, sub_type: Optional[type] = None,
                                        allow_additions: bool = True) -> QWidget:
        widget = IterableEditWidget(self, sub_type, allow_additions)
        if option != sub_type:
            for entry in option:
                sub_widget, getter = self.construct_widget_based_on_type(entry)
                widget.add(sub_widget, getter)
        return widget

    @staticmethod
    def __generate_checkbox_widget(option: bool) -> QCheckBox:
        check_box = QCheckBox()
        check_box.setChecked(option)
        return check_box

    @staticmethod
    def __generate_selection_box(options: List[str]) -> NoWheelComboBox:
        box = NoWheelComboBox()
        box.addItems(options)
        return box

    @staticmethod
    def __generate_line_editor_widget(option: Union[str, List]) -> QLineEdit:
        line_edit = QLineEdit()
        line_edit.setText(str(option))
        return line_edit

    def __generate_removal_widget(self, index: int) -> QPushButton:
        check_box = QPushButton(icon=self.style().standardIcon(QStyle.SP_TitleBarCloseButton), text="remove field")
        check_box.setChecked(False)
        check_box.clicked.connect(lambda: self.__remove_option(index))  # pylint: disable=no-member
        return check_box

    def __remove_option(self, row_index: int):
        to_delete = []
        # this method can be reached even if removal is not allowed, because we want to remove things
        # if we are loading the new options and they have duplicate names to existing ones
        # the reason is, because we only hold the widgets and getters, but no setters
        n_cols = 3 if self._allow_removal else 2
        for i in range(n_cols):
            item = self.__layout.itemAtPosition(row_index, i)
            assert item is not None
            to_delete.append(item.widget())
        assert isinstance(to_delete[0], QLabel)
        name = to_delete[0].text()
        del self._options[name]
        del self._option_widgets[name]
        for widget in to_delete:
            widget.setAttribute(Qt.WA_DeleteOnClose)
            self.__layout.removeWidget(widget)
            widget.close()
            widget.setParent(None)  # type: ignore
        self.__layout.update()
        self.updateGeometry()

    @staticmethod
    def get_attributes_of_object(o: object) -> Dict[str, Any]:
        attributes = inspect.getmembers(
            o.__class__, lambda a: not (inspect.isroutine(a))
        )
        return {
            a[0]: getattr(o, a[0])
            for a in attributes
            if a[0] != "step_result" and not a[0].startswith("_") and not inspect.isclass(a[1])
        }

    @staticmethod
    def set_attributes_to_object(o: object, d: Dict[str, Any]) -> None:
        if isinstance(o, dict) or isinstance(o, ValueCollection):
            for option_name, option in d.items():
                o[option_name] = option
        else:
            for option_name, option in d.items():
                setattr(o, option_name, option)

    def get_widget_data(self) -> Dict[str, Any]:
        data: Dict[str, Any] = dict()
        for option_name, option in self._options.items():
            if option_name in self._ignore_list:
                continue
            try:
                getter = self._option_getters[option_name]
                data[option_name] = getter()
            except BaseException as e:
                write_error_message(f"Could not get data for option {option_name}: {e}")
                widget = self._option_widgets[option_name]
                try:
                    sub_data = widget.get_widget_data()
                    attr = self.get_attributes_of_object(option)
                    if attr:
                        set_attr = {data_k: data_v for data_k, data_v in sub_data.items() if data_k in attr}
                        self.set_attributes_to_object(option, set_attr)
                    if hasattr(option, "options"):
                        attr = self.get_attributes_of_object(option.options)
                        set_attr = {data_k: data_v for data_k, data_v in sub_data.items() if data_k in attr}
                        self.set_attributes_to_object(option.options, set_attr)
                    data[option_name] = option
                except BaseException as exc:  # TODO maybe remove exception info for release
                    option_info = option if option is type else type(option)
                    raise NotImplementedError(f"Type '{str(option_info)}' is not implemented: {str(exc)}") from exc
        return data

    def _save(self):
        filename = get_save_file_name(self, "options", self.__io_extensions)
        if filename is None:
            return
        data = self.get_widget_data()
        # delete excluded keys
        if self._keys_excluded_from_io is not None:
            for key in self._keys_excluded_from_io:
                data.pop(key, None)
        try:
            if filename.suffix == ".json":
                with open(filename, "w") as f:
                    f.write(encode(data))
            elif filename.suffix in [".pickle", ".pkl"]:
                with open(filename, "wb") as f:
                    pickle.dump(data, f)
            elif filename.suffix == ".yaml":
                with open(filename, "w") as f:
                    f.write(yaml.dump(data))
            else:
                raise NotImplementedError(f"Extension {filename.suffix} is not implemented")
        except BaseException as e:
            write_error_message(f"Could not save options: {e}")

    def _load(self):
        filename = get_load_file_name(self, "options", self.__io_extensions)
        if filename is None:
            return
        try:
            if filename.suffix == ".json":
                with open(filename, "r") as f:
                    data = decode(f.read())
            elif filename.suffix in [".pickle", ".pkl"]:
                with open(filename, "rb") as f:
                    data = pickle.load(f)
            elif filename.suffix == ".yaml":
                with open(filename, "r") as f:
                    data = yaml.safe_load(f.read().replace("\t", "  ").replace("    ", "  "))
            else:
                raise NotImplementedError(f"Extension {filename.suffix} is not implemented")
        except BaseException as e:
            write_error_message(f"Could not save options: {e}")
            return
        # delete excluded keys
        if self._keys_excluded_from_io is not None:
            for key in self._keys_excluded_from_io:
                data.pop(key, None)
        # delete existing options with the same name
        # we need to first gather the indices and then delete from highest to lowest, because otherwise
        # Qt gives us the wrong indices
        to_delete_rows = []
        for option_name, option in data.items():
            if option_name in self._options:
                index = self.__layout.indexOf(self._option_widgets[option_name])
                # index is the total index of all widgets in the layout, we want to know the row
                # index is always the value position in the row, which has the index 1 with 0 being the label
                # therefore we remove 1 and the divide by the number of columns
                n_cols = 3 if self._allow_removal else 2
                row = (index - 1) // n_cols
                to_delete_rows.append(row + 1)
        to_delete_rows.sort(reverse=True)
        for row in to_delete_rows:
            self.__remove_option(row)
        # now add the new options
        for option_name, option in data.items():
            self.add_key_value(option_name, option)


class OptionsAdditionDialog(QDialog):
    """
    QDialog to add an option
    """

    def __init__(self, add_value_field: bool, parent: Optional[QWidget] = None,
                 suggestions: Optional[List[str]] = None) -> None:
        super().__init__(parent)

        self.setWindowTitle("Specify new key name")
        self.setMinimumWidth(400)
        self.setMinimumHeight(400)
        self.setMaximumWidth(650)
        self.setMaximumHeight(900)

        self.name = ""
        self.type = ""
        self.__add_value_field = add_value_field

        layout = QVBoxLayout()
        name_label = QLabel("Option name")
        self.__name_edit = QLineEdit()
        self.__name_edit.setText("")
        if suggestions is not None:
            completer = QCompleter(suggestions)
            completer.setFilterMode(Qt.MatchContains)
            self.__name_edit.setCompleter(completer)
        layout.addWidget(name_label)
        layout.addWidget(self.__name_edit)

        if add_value_field:
            type_label = QLabel("Option type (e.g. 'str')")
            type_suggestions = ['str', 'int', 'float', 'bool', 'list', 'dict', 'set', 'tuple']
            type_completer = QCompleter(type_suggestions)
            self.__type_edit = QLineEdit()
            self.__type_edit.setText("")
            self.__type_edit.setCompleter(type_completer)
            layout.addWidget(type_label)
            layout.addWidget(self.__type_edit)

        layout.addWidget(TextPushButton("Ok", self.reject, self))

        self.setLayout(layout)

    def closeEvent(self, event: QCloseEvent) -> None:
        self.name = self.__name_edit.text()
        if self.__add_value_field:
            self.type = self.__type_edit.text()
        super().closeEvent(event)

    def reject(self) -> None:
        self.name = self.__name_edit.text()
        if self.__add_value_field:
            self.type = self.__type_edit.text()
        super().reject()


class IterableEditWidget(QWidget):

    def __init__(self, parent: DictOptionWidget, fixed_sub_types: Union[type, List[type], None] = None,
                 allow_additions: bool = False) -> None:
        super().__init__(parent)
        self._parent = parent
        # buttons
        self._add_button = None
        self._copy_button = None
        self._paste_button = None
        self.remove_last_button = None
        # containers
        self._widgets: List[QWidget] = []
        self._getters: List[Callable[[], Any]] = []
        self._super_layout = VerticalLayout()
        self.single_layout = QHBoxLayout()
        self._super_layout.addLayout(self.single_layout)
        self.setLayout(self._super_layout)
        # find out initial structure
        if fixed_sub_types == Any:
            self._fixed_sub_types = None
        else:
            self._fixed_sub_types = fixed_sub_types
        self._add_wrapper(self._fixed_sub_types, add_one=False)
        self._allow_additions = allow_additions
        self._copy_filename = ".iterable_edit_widget_copy.pkl"
        if allow_additions:
            # add buttons
            max_width = 200
            self._add_button = TextPushButton("Add", self._add_dialog, max_width=max_width)
            self._copy_button = TextPushButton("Copy", self.copy, max_width=max_width)
            self._paste_button = TextPushButton("Paste", self.paste, max_width=max_width)
            self._super_layout.add_widgets([self._add_button, self._copy_button, self._paste_button])
            if not self._widgets or self.single_layout.count():
                self.remove_last_button = TextPushButton("Remove last", self._remove_last_single_entry)
                self.single_layout.addWidget(self.remove_last_button)

    def copy(self):
        """
        write a hidden pickle file
        """
        obj = [self._fixed_sub_types, self._allow_additions, self.get_values()]
        with open(self._copy_filename, "wb") as f:
            pickle.dump(obj, f)

    def paste(self):
        """
        read a hidden pickle file
        """
        if not path.isfile(self._copy_filename):
            write_error_message("Cannot paste, No copy available")
            return
        with open(self._copy_filename, "rb") as f:
            copy = pickle.load(f)
        if not isinstance(copy, list) or len(copy) != 3:
            write_error_message("Cannot paste, copy is not an IterableEditWidget")
            return
        self.clear()
        self._fixed_sub_types = copy[0]
        self._allow_additions = copy[1]
        for value in copy[2]:
            self._add_by_value(value)

    def clear(self):
        """
        clear all widgets
        """
        for widget in self._widgets:
            if any(self.single_layout.itemAt(i).widget() == widget for i in range(self.single_layout.count())):
                self._remove_entries(widget, self.single_layout)
            elif any(self._super_layout.itemAt(i).widget() == widget for i in range(self._super_layout.count())):
                self._remove_entries(widget, self._super_layout)
            else:
                pass

    def _add_by_value(self, value: Any) -> None:
        widget, getter = self._parent.construct_widget_based_on_type(value)
        self.add(widget, getter)

    def add(self, other: QWidget, getter: Callable[[], Any]) -> None:
        self._widgets.append(other)
        self._getters.append(getter)
        if not isinstance(other, IterableEditWidget):
            # single remove button
            offset = 0 if self.remove_last_button is None else 1
            self.single_layout.insertWidget(self.single_layout.count() - offset, other)
        elif self._add_button is None:
            self._super_layout.addWidget(other)
        else:
            remove_button = QPushButton("Remove")
            remove_button.clicked.connect(  # pylint: disable=no-member
                lambda: self._remove_entries(other, self._super_layout)
            )
            # use inserts because of buttons
            offset = 0 if other.remove_last_button is None else 1
            other.single_layout.insertWidget(other.single_layout.count() - offset, remove_button)
            # add, copy, paste buttons
            super_offset = 0 if self._add_button is None else 3
            self._super_layout.insertWidget(self._super_layout.count() - super_offset, other)

    def _remove_last_single_entry(self) -> None:
        if not self._widgets:
            return
        offset = 1 if self._add_button is None else 2  # because of remove button
        item = self.single_layout.itemAt(self.single_layout.count() - offset)
        self._remove_entries(item.widget(), self.single_layout)

    def _remove_entries(self, widget: QWidget, layout) -> None:
        # our data structures
        index = self._widgets.index(widget)
        self._widgets.pop(index)
        self._getters.pop(index)
        # widget visibility
        widget.setAttribute(Qt.WA_DeleteOnClose)
        widget.close()
        widget.setParent(None)  # type: ignore
        layout.removeWidget(widget)

    def get_values(self) -> Iterable:
        return [getter() for getter in self._getters]

    def _add_dialog(self) -> None:
        if self._fixed_sub_types is None and not self._getters:
            # no type specified, ask for type
            text = text_input_box(self, "Add list entry", "Please specify the type of the list entry",
                                  QLineEdit.Normal, "int")
            if not text:
                write_error_message("No type specified, aborting")
                return
            try:
                cls = eval(text)  # pylint: disable=eval-used
            except NameError:
                write_error_message(f"Given type {text} is not a valid type")
                return
            except TypeError:
                write_info_message(f"Trying to infer type from {text}")
                try:
                    cls = type(text)
                except TypeError:
                    write_error_message(f"Could not infer type from {text}")
                    return
            self._fixed_sub_types = cls
        elif self._fixed_sub_types is None:
            self._fixed_sub_types = type(self._getters[-1]())
        self._add_wrapper(self._fixed_sub_types, add_one=True)

    def _add_wrapper(self, sub_types: Union[type, List[type], None], add_one: bool = False) -> None:
        if isinstance(sub_types, list):
            for t in sub_types:
                widget, getter = self._parent.construct_widget_based_on_type(t)
                self.add(widget, getter)
        elif sub_types is not None and hasattr(sub_types, "__origin__") \
                and getattr(sub_types, "__origin__") is not Union:
            widget = IterableEditWidget(self._parent, list(getattr(sub_types, "__args__", True)))  # type: ignore
            self.add(widget, widget.get_values)
        elif add_one:
            widget, getter = self._parent.construct_widget_based_on_type(sub_types)
            self.add(widget, getter)

    def __len__(self) -> int:
        return len(self._widgets)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from PySide2.QtWidgets import QWidget
from PySide2.QtGui import QCursor
from PySide2.QtCore import Qt


class Clickable(QWidget):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))

    def setEnabled(self, value: bool) -> None:
        cursor = Qt.CursorShape.PointingHandCursor if value else Qt.CursorShape.ForbiddenCursor
        self.setCursor(QCursor(cursor))
        super().setEnabled(value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import Any, List, Sequence, Tuple, Union

from PySide2.QtWidgets import QTreeWidget, QTreeWidgetItem, QMenu, QSizePolicy, QWidget

from scine_utilities import AtomCollection

import scine_heron.config as config
from scine_heron.styling.delegates import CustomLightDelegate
from scine_heron.molecular_viewer import get_mol_viewer_tab
from scine_heron.utilities import write_info_message, copy_text_to_clipboard


class TreeWidget(QTreeWidget):
    """
    Wrapper around the standard QTreeWidget that holds a TreeWidgetItem and
    propagates some camera information and focus actions.
    """

    def __init__(self, parent: QWidget):
        from .current_selection import SelectionTab
        super().__init__(parent)
        self._parent: SelectionTab = parent
        self.itemClicked.connect(self.focus)  # pylint: disable=no-member
        self.itemActivated.connect(self.focus)  # pylint: disable=no-member
        self.setMinimumHeight(250)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.updateGeometry()
        if config.MODE in config.LIGHT_MODES:
            self.setItemDelegate(CustomLightDelegate())

    def insertTopLevelItems(self, index: int, items: Sequence) -> None:
        super().insertTopLevelItems(index, items)
        self._parent.reset_camera()

    def contextMenuEvent(self, event):
        self.currentItem().contextMenuEvent(event)

    def focus(self, item: QTreeWidgetItem, _: Any) -> None:
        self._parent.display_atoms_with_sites(item.atoms, item.indices)


class TreeWidgetItem(QTreeWidgetItem):
    """
    A wrapper around the standard QTreeWidgetItem that holds the name, a structure, and reactive indices.
    It is meant to be plugged into the TreeWidget.
    """

    def __init__(self, parent: Union[TreeWidget, QTreeWidgetItem], info: List[str], atoms: AtomCollection,
                 indices: List[int]) -> None:
        """
        Construct the item

        Parameters
        ----------
        parent : Union[TreeWidget, QTreeWidgetItem]
            Should be the TreeWidget or another TreeWidgetItem, not generic
            (not possible currently because of backwards compatibility)
        info : List[str]
            The information to be displayed in the tree, from which the type and name are extracted
        atoms : AtomCollection
            The structure
        indices : List[int]
            The reactive sites of the structure
        """
        super().__init__(parent, info, type=QTreeWidgetItem.UserType)
        self._parent = parent
        assert len(info) == 1
        self.item_type = info[0].split('(')[0].strip()
        id_string = info[0].split('(')[1][:-1]
        self.names: Tuple[str, ...] = tuple([s.strip() for s in id_string.split("-")])
        self.atoms = atoms
        self.indices = indices

    def contextMenuEvent(self, event):
        menu = QMenu()
        show_in_mol_view = menu.addAction('Move to Interactive')
        show_all_reactive_sites = menu.addAction('Display all reactive sites')
        copy_names = menu.addAction('Copy ID')
        _ = menu.addAction('Close Menu')
        action = menu.exec_(event.globalPos())  # type: ignore
        if action == show_in_mol_view:
            tab = get_mol_viewer_tab(want_atoms_there=False)
            if tab is not None:
                tab.update_molecule(atoms=self.atoms)
                tab.mol_widget.set_selection(self.indices)
        elif action == show_all_reactive_sites:
            self.show_all_sites()
        elif action == copy_names:
            copy_text_to_clipboard("-".join(self.names))

    def show_all_sites(self):
        """
        Combine all reactive sites of individual complexes and display them all at once (with deduplication).
        This is useful to get a quick overview of the reactive sites of a structure to possibly adapt the
        selection.
        """
        if not self.indices:
            write_info_message("No reactive sites")
            return
        if "Structure" in self.item_type:
            self._parent.show_all_sites()
        else:
            # none structure item should hold all indices as indices
            self._parent.focus(self, None)

    def focus(self, item: QTreeWidgetItem, _: Any) -> None:
        self._parent.focus(item, _)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

import os.path
import sys
import argparse
import pathlib
import signal
import typing

from PySide2.QtWidgets import QApplication
from PySide2.QtGui import QIcon
from qt_material import apply_stylesheet, get_theme

from .main_window import MainWindow
from scine_heron.resources import resource_path
import scine_heron.config as config


def parse_command_line_arguments() -> argparse.Namespace:
    """
    Parses the command line arguments and returns them.
    """
    parser = argparse.ArgumentParser(description="SCINE User Interface")
    parser.add_argument(
        "-f", "--file", help="load molecule from FILE", type=pathlib.Path, default=None,
    )
    parser.add_argument(
        "-mf", "--method-family", help="Specify method family of calculator", type=str, default=None,
    )
    parser.add_argument(
        "-pr", "--program", help="Specify program of calculator", type=str, default=None,
    )
    parser.add_argument(
        "-p", "--port", help="SCINE database port", type=int, default=27017,
    )
    parser.add_argument(
        "-n",
        "--name",
        help="SCINE database name",
        type=str,
        default="default",
    )
    parser.add_argument(
        "-i",
        "--ip",
        help="SCINE database IP or hostname",
        type=str,
        default="localhost",
    )
    parser.add_argument(
        "-c",
        "--connect",
        help="Immediately connect with SCINE database",
        action='store_true',  # switch flip with argument only
    )
    parser.add_argument(
        "-m", "--mode", help="choose dark or light mode", type=str, default="dark"
    )
    return parser.parse_args()


def main() -> None:
    """
    Starts the molecule viewer application.
    """
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    application = QApplication(sys.argv)
    application.setWindowIcon(QIcon(os.path.join(resource_path(), 'heron_logo.png')))

    arguments = parse_command_line_arguments()
    theme_path = os.path.join(resource_path(), f'theme_{arguments.mode}.xml')
    css_path = os.path.join(resource_path(), f'theme_{arguments.mode}.css')
    if os.path.exists(theme_path):
        apply_stylesheet(
            application,
            theme=theme_path,
            css_file=css_path,
            invert_secondary=True,
        )
        config.COLORS = get_theme(theme_path)
        config.MODE = arguments.mode
    else:
        raise RuntimeError(f'Failed to load theme at "{theme_path}"')
    filename = typing.cast(typing.Optional[pathlib.Path], arguments.file)
    # Avoid segfault in QFontEngineFT::loadGlyphSet
    # Solution according to:
    # https://stackoverflow.com/questions/40490414/cannot-trace-cause-of-crash-in-qt-program
    os.environ['QT_NO_FT_CACHE'] = "1"
    window = MainWindow(
        file_name=filename,
        method_family=arguments.method_family,
        program=arguments.program,
        db_credentials=(arguments.name, arguments.ip, arguments.port),
        connect_with_db=arguments.connect
    )
    window.show()  # type: ignore[no-untyped-call]
    application.exec_()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Functions to create and work with colorbars.
"""

import vtk
import typing
import sys


def create_symmetric_color_transfer_function(
    value_range: typing.Tuple[float, float]
) -> vtk.vtkColorTransferFunction:
    """
    Creates a default symmetric vtkColorTransferFunction with the given range.
    In particular, the color white is used for the value 0.
    """
    assert value_range[0] <= value_range[1]

    series = vtk.vtkColorSeries()
    series.SetColorScheme(vtk.vtkColorSeries.BREWER_DIVERGING_BROWN_BLUE_GREEN_11)

    # The colors need to be converted from range [0, 255] to range [0, 1].
    colors = [
        (x / 255 for x in series.GetColor(i)) for i in range(series.GetNumberOfColors())
    ]

    # Spread points evenly over [-m, m], where m = max(|x| for x in value_range)
    # (note that this means that the 5th color (white) is assigned to 0).
    #
    # In case the values are (almost) zero, we use a range of [-eps, eps] to avoid
    # identical points.
    absolute_max = max(max(abs(x) for x in value_range), sys.float_info.epsilon)
    points = [
        (2 * i / (len(colors) - 1) - 1) * absolute_max for i in range(len(colors))
    ]

    function = vtk.vtkColorTransferFunction()

    # Reverse colors to assign blue to negative numbers.
    for point, color in zip(points, reversed(colors)):
        function.AddRGBPoint(point, *color)

    # The colorbar only needs to show used colors.
    function.AdjustRange(value_range)

    return function


def create_colorbar() -> vtk.vtkScalarBarActor:
    """
    Creates a default color bar with default colors.
    """
    colorbar = vtk.vtkScalarBarActor()
    colorbar.SetMaximumWidthInPixels(50)
    colorbar.GetLabelTextProperty().SetBold(False)
    colorbar.GetLabelTextProperty().SetItalic(False)

    from scine_heron.utilities import qcolor_by_key
    color = qcolor_by_key('primaryTextColor')
    colorbar.GetLabelTextProperty().SetColor(color.getRgbF()[0], color.getRgbF()[1], color.getRgbF()[2])

    colorbar.GetLabelTextProperty().ShadowOff()
    return colorbar



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the MoleculeInteractorStyle class.
"""

from scine_heron.molecule.animator import Animator
from scine_heron.status_manager import StatusManager
from scine_heron.energy_profile.energy_profile_status_manager import (
    EnergyProfileStatusManager,
)
from scine_heron.haptic.haptic_client import HapticClient
from scine_heron.electronic_data.electronic_data_status_manager import (
    ElectronicDataStatusManager,
)
from scine_heron.settings.settings_status_manager import SettingsStatusManager
from scine_heron.molecule.styles.haptic_interactor_style import HapticInteractorStyle
from scine_heron.molecule.create_molecule_animator import create_molecule_animator
from uuid import uuid1
from typing import Optional, Dict, Callable, List

from vtk import (
    vtkRenderer,
    vtkMoleculeMapper,
    vtkMolecule,
    vtkActor,
)
from vtkmodules.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor


class MoleculeInteractorStyle(HapticInteractorStyle):
    """
    An interactor style (based on "HapticInteractorStyle") that supports
    picking and moving particles with mouse or haptic device.
    """

    def __init__(
        self,
        interactor: QVTKRenderWindowInteractor,
        renderer: vtkRenderer,
        mapper: vtkMoleculeMapper,
        haptic_client: Optional[HapticClient],
        actors_dict: Dict[str, vtkActor],
        selected_atom_callback: Callable[[Optional[List[int]]], None],
    ):
        super(MoleculeInteractorStyle, self).__init__(
            interactor,
            renderer,
            mapper,
            haptic_client,
            actors_dict,
            selected_atom_callback,
        )

        self.__haptic_client = haptic_client
        self.__animator: Optional[Animator] = None

        self.__energy_status_manager: Optional[EnergyProfileStatusManager] = None
        self.__charge_status_manager: StatusManager[Optional[List[float]]] = None  # type: ignore[assignment]
        self.__force_status_manager: StatusManager[Optional[List[float]]] = None  # type: ignore[assignment]

    def set_status_managers(
        self,
        settings_status_manager: SettingsStatusManager,
        energy_status_manager: Optional[EnergyProfileStatusManager],
        charge_status_manager: StatusManager[Optional[List[float]]],
        force_status_manager: StatusManager[Optional[List[float]]],
        electronic_data_status_manager: ElectronicDataStatusManager,
    ) -> None:
        super().set_settings_status_manager(settings_status_manager)
        super().set_electronic_data_status_manager(electronic_data_status_manager)

        self.__energy_status_manager = energy_status_manager
        self.__charge_status_manager = charge_status_manager
        self.__force_status_manager = force_status_manager

    def set_calc_gradient_in_loop(self, calc_gradient_in_loop: bool) -> None:
        """
        Save calc_gradient_in_loop flag and start or stop the loop.
        """
        if self.__haptic_client is not None:
            self.__haptic_client.set_calc_gradient_in_loop(calc_gradient_in_loop)

        if self.__animator is not None:
            if calc_gradient_in_loop:
                self.__animator.start()
            else:
                self.__animator.stop()

    def __setup_animator(self) -> None:
        """
        Create molecule gradient calculator.
        """

        old_animator_running = False
        if self.__animator is not None:
            # Cleanup of current animator
            old_animator_running = self.__animator.running
            self.__animator.render_signal.disconnect(self._render)
            self.__animator.stop()

        self.__animator = create_molecule_animator(
            self._molecule_version.int,
            self._molecule,
            self._settings_status_manager,
            self.__haptic_client,
            self.__energy_status_manager,
            self._electronic_data_status_manager,
            self.__charge_status_manager,
            self.__force_status_manager,
        )

        self.__animator.render_signal.connect(self._render)

        # If the old animator was running before,
        # start the new one
        if old_animator_running:
            self.__animator.start()

    @property
    def molecule(self) -> vtkMolecule:
        return self._molecule

    @molecule.setter
    def molecule(self, molecule: vtkMolecule) -> None:
        self._molecule = molecule
        self._molecule_version = uuid1()
        self.__setup_animator()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
from scine_heron.database.runtime_histogram_dialog import RuntimeHistogramDialog
from scine_heron.utilities import (
    color_axis,
    color_figure,
    get_font,
)

from PySide2.QtWidgets import QWidget, QPushButton, QGridLayout, QFrame
from PySide2.QtGui import QKeySequence

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure
from matplotlib.colors import to_hex, to_rgb
from json import dumps
import numpy as np
from typing import List, Optional
from scine_database import Manager


class TwoPieCharts(FigureCanvasQTAgg):
    def __init__(self, width=5, height=4):
        self.fig = Figure(figsize=(width, height))
        self.ax1 = self.fig.add_subplot(1, 2, 1)
        self.ax2 = self.fig.add_subplot(1, 2, 2)
        color_figure(self.fig)
        self.font = get_font()
        color_axis(self.ax1)
        color_axis(self.ax2)
        super(TwoPieCharts, self).__init__(self.fig)

    def update_statistics(self, db_manager: Manager, fake: bool = False) -> None:

        select_all = dumps({})
        calculations = db_manager.get_collection("calculations")
        structures = db_manager.get_collection("structures")
        compounds = db_manager.get_collection("compounds")
        flasks = db_manager.get_collection("flasks")
        properties = db_manager.get_collection("properties")
        elementary_steps = db_manager.get_collection("elementary_steps")
        reactions = db_manager.get_collection("reactions")

        element_counts_labels = [
            "Calculations",
            "Structures",
            "Compounds",
            "Flasks",
            "Properties",
            "Elementary Steps",
            "Reactions",
        ]
        if fake:
            element_counts_data = [0, 0, 0, 0, 0, 0, 0]
        else:
            element_counts_data = [
                calculations.count(select_all),
                structures.count(select_all),
                compounds.count(select_all),
                flasks.count(select_all),
                properties.count(select_all),
                elementary_steps.count(select_all),
                reactions.count(select_all),
            ]
        self.set_up_pie(
            "ax1", "Document Counts", element_counts_data, element_counts_labels
        )

        calculation_counts_labels = ["Hold", "New", "Pending", "Complete", "Failed"]
        if fake:
            calculation_counts_data = [0, 0, 0, 0, 0]
        else:
            calculation_counts_data = [
                calculations.count(dumps({"status": "hold"})),
                calculations.count(dumps({"status": "new"})),
                calculations.count(dumps({"status": "pending"})),
                calculations.count(dumps({"status": "complete"})),
                calculations.count(dumps({"status": "failed"})),
            ]
        self.set_up_pie(
            "ax2",
            "Calculation Counts",
            calculation_counts_data,
            calculation_counts_labels,
        )

        self.fig.tight_layout()
        self.fig.canvas.draw_idle()

    def set_up_pie(
        self, ax: str, title: str, data: List[int], labels: List[str]
    ) -> None:
        axis = getattr(self, ax)
        axis.cla()

        def color_fade(c1, c2, mix):
            c1 = np.array(to_rgb(c1))
            c2 = np.array(to_rgb(c2))
            return to_hex((1 - mix) * c1 + mix * c2)

        colors = [
            color_fade("#214478", "#AFC6E9", i / len(labels))
            for i in range(len(labels))
        ]

        if sum(data) == 0:
            pie_data = [1 for _ in labels]
        else:
            pie_data = data

        pie = axis.pie(
            pie_data,
            colors=colors,
            explode=[0 for _ in labels],
            rotatelabels=True,
            shadow=False,
            startangle=90,
        )
        axis.set_title(title, self.font)

        if sum(data) == 0.0:
            percentages = [0.0 for _ in pie_data]
        else:
            percentages = [i * 100 / sum(pie_data) for i in pie_data]

        ll = [
            "{:d} ({:4.2f} %) {:s}".format(b, p, a)
            for a, b, p in zip(labels, data, percentages)
        ]
        axis.legend(pie[0], ll, loc="upper center", bbox_to_anchor=(0.5, -0.05))
        axis.axis("equal")


class QHLine(QFrame):
    def __init__(self):
        super(QHLine, self).__init__()
        self.setFrameShape(QFrame.HLine)
        self.setFrameShadow(QFrame.Sunken)


class DatabaseMonitorWidget(QWidget):
    def __init__(self, parent: Optional[QWidget], db_manager) -> None:
        super(DatabaseMonitorWidget, self).__init__(parent)
        self.db_manager = db_manager

        # Create layout and add widgets
        layout = QGridLayout()
        self.element_counts = TwoPieCharts()
        self.element_counts.update_statistics(self.db_manager, fake=True)
        layout.addWidget(self.element_counts, 0, 0, 1, 3)
        self.button_update = QPushButton("Update")
        self.button_update.setShortcut(QKeySequence("r"))
        layout.addWidget(self.button_update, 1, 1)
        self.button_update.clicked.connect(self.update_statistics)  # pylint: disable=no-member
        layout.addWidget(QHLine(), 2, 0, 1, 3)

        self.button_runtime = QPushButton("Runtime Histogram")
        layout.addWidget(self.button_runtime, 3, 1)
        self.button_runtime.clicked.connect(self.display_runtime_histogram)  # pylint: disable=no-member

        # Set dialog layout
        self.setLayout(layout)

    def update_statistics(self) -> None:
        self.button_update.setText("Updating...")
        self.element_counts.update_statistics(self.db_manager)
        self.button_update.setText("Update")

    def display_runtime_histogram(self) -> None:
        RuntimeHistogramDialog(self, self.db_manager)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""


from PySide2.QtWidgets import QGridLayout, QPushButton, QWidget

from scine_heron.autocas.signal_handler import SignalHandler


class AutocasController(QWidget):
    def __init__(self, parent: QWidget, signal_handler: SignalHandler):
        QWidget.__init__(self, parent)

        self.signal_handler = signal_handler

        self.__layout = QGridLayout()

        self.button_generate_initial_orbitals = QPushButton("Generate Initial Orbitals")
        self.__layout.addWidget(self.button_generate_initial_orbitals)
        # pylint: disable-next=E1101
        self.button_generate_initial_orbitals.clicked.connect(
            self.signal_handler.start_initial_orbital_calculation_signal
        )

        self.button_run_initial_dmrg = QPushButton("Run Initial DMRG")
        self.__layout.addWidget(self.button_run_initial_dmrg)
        # pylint: disable-next=E1101
        self.button_run_initial_dmrg.clicked.connect(
            self.signal_handler.start_initial_dmrg_calculation_signal
        )

        self.button_run_final_calculation = QPushButton("Run Final Calculation")
        self.__layout.addWidget(self.button_run_final_calculation)
        # pylint: disable-next=E1101
        self.button_run_final_calculation.clicked.connect(
            self.signal_handler.start_initial_dmrg_calculation_signal
        )

        self.button_open_entanglement = QPushButton("Show Entganglement")
        self.__layout.addWidget(self.button_open_entanglement)
        # pylint: disable-next=E1101
        self.button_open_entanglement.clicked.connect(self.open_entanglement)
        # pylint: disable-next=E1101
        self.button_open_entanglement.clicked.connect(
            self.signal_handler.open_entanglement_widget_signal
        )

        self.button_open_threshold = QPushButton("Show Threshold")
        self.__layout.addWidget(self.button_open_threshold)
        # pylint: disable-next=E1101
        self.button_open_threshold.clicked.connect(self.open_threshold)

        self.button_open_orbitals = QPushButton("Show Orbitals")
        self.__layout.addWidget(self.button_open_orbitals)
        # pylint: disable-next=E1101
        self.button_open_orbitals.clicked.connect(self.open_orbitals)
        # pylint: disable-next=E1101
        self.button_open_orbitals.clicked.connect(
            self.signal_handler.open_molecule_widget_signal
        )

        self.setLayout(self.__layout)

    # @Slot()

    def open_entanglement(self):
        pass

    def open_threshold(self):
        pass

    def open_orbitals(self):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Utility functions for working with VTK arrays.
"""

import typing
import sys
import math
from itertools import islice, repeat, chain
import vtk


def iterable_to_vtk_array(
    data: typing.Iterable[float], number_of_values: int, default_value: float = 0.0
) -> vtk.vtkDoubleArray:
    """
    Stores the numbers generated by the iterable to a new
    vtkDoubleArray and returns this array.

    The parameter `number_of_values` is used to pre-allocate
    the correct size.
    """

    result = vtk.vtkDoubleArray()
    result.SetNumberOfValues(number_of_values)
    for index, value in enumerate(
        islice(chain(data, repeat(default_value)), number_of_values)
    ):
        result.SetValue(index, value)
    return result


def rescale_to_range(
    array: vtk.vtkDoubleArray,
    to_range: typing.Tuple[float, float],
    rel_tol: float = sys.float_info.epsilon,
    abs_tol: float = sys.float_info.epsilon,
) -> vtk.vtkDoubleArray:
    """
    Rescales the given array to the desired range.

    The tolerances are used to determine whether the input range has span zero.
    In this case a small nonzero span is used.
    """
    to_span = to_range[1] - to_range[0]
    assert to_span >= 0

    # The values need to span a positive range to be able to scale to `to_range`.
    # We use at least a small span derived from the tolerances.
    array_range = array.GetValueRange()
    array_span = array_range[1] - array_range[0]
    array_center = array_range[0] + array_span / 2
    from_range = (
        array_range
        if not math.isclose(array_span, 0.0, rel_tol=rel_tol, abs_tol=abs_tol)
        else (
            array_center - max(rel_tol * abs(array_center), abs_tol),
            array_center + max(rel_tol * abs(array_center), abs_tol),
        )
    )
    from_span = from_range[1] - from_range[0]

    assert not math.isclose(from_span, 0.0, rel_tol=rel_tol, abs_tol=abs_tol)
    factor = to_span / from_span

    result = vtk.vtkDoubleArray()
    result.SetNumberOfValues(array.GetNumberOfValues())
    for id in range(array.GetNumberOfValues()):
        result.InsertValue(
            id, to_range[0] + (array.GetValue(id) - from_range[0]) * factor
        )

    return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""

from typing import Dict

COLORS: Dict[str, str] = {
    'primaryColor': '#94bbff',
    'primaryLightColor': '#94bbff',
    'secondaryColor': '#1e304f',
    'secondaryLightColor': '#061e3b',
    'secondaryDarkColor': '#4a597b',
    'primaryTextColor': '#ffffff',
    'secondaryTextColor': '#ffffff',
    'compoundColor': '#8dd35f',
    'reactionColor': '#ff9955',
    'associationColor': '#d4aa00',
    'flaskColor': '#5f8dd3',
    'structureColor': '#154734',
    'elementaryStepColor': '#a11f05',
    'borderColor': '#000000',
    'edgeColor': '#c8c8c8',
    'highlightColor': '#d40000',
}

MODE = 'dark'

LIGHT_MODES = ['light', 'legacy']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the ElectronicData class.
"""
import re
import math
import numpy as np
from typing import List, Any
from scine_heron.molecule.utils.molecule_utils import times_angstrom_per_bohr


class MolecularOrbital:
    """
    Provide molecular orbital.
    """

    def __init__(
        self,
        symmetry: str,
        energy: float,
        spin: str,
        occupation: float,
        coefficients: List[float],
    ):
        self.symmetry = symmetry
        self.energy = energy
        self.spin = spin
        self.occupation = occupation
        self.coefficients = np.array(coefficients)

    @classmethod
    def from_molden_file(cls, lines: List[str]) -> Any:
        symmetry = re.match(r"\s?Sym(?:.*)=(.*)", lines[0]).groups()[0].strip()  # type: ignore[union-attr]
        energy = float(
            re.match(  # type: ignore[union-attr]
                r"\s?Ene(?:.*)=(.*)",
                lines[1]).groups()[0].strip())  # type: ignore[union-attr]
        spin = re.match(r"\s?Spin(?:.*)=(.*)", lines[2]).groups()[0].strip()  # type: ignore[union-attr]
        occupation = float(
            re.match(  # type: ignore[union-attr]
                r"\s?Occup(?:.*)=(.*)",
                lines[3]).groups()[0].strip())  # type: ignore[union-attr]
        coefficients = [float(line.split()[1]) for line in lines[4:] if len(line) > 0]

        return cls(symmetry, energy, spin, occupation, coefficients)


class GaussianOrbital:
    """
    Provide gaussian orbital.
    """

    def __init__(self, orb_type: str, coefficients: List[List[float]]):
        self.orb_type = orb_type
        self.nr_gaussians = len(coefficients)

        self.coefficients = coefficients
        self.alpha = np.array([coefficients[i][0] for i in range(self.nr_gaussians)])
        self.coeff = np.array(
            [
                self.__calculate_coefficients(i, orb_type)
                for i in range(self.nr_gaussians)
            ]
        )

    def chi_step(self) -> int:
        if self.orb_type == "s":  # s orbital
            return 1
        elif self.orb_type == "p":  # p orbital
            return 3
        elif self.orb_type == "d":  # d orbital [5D]
            return 5
        elif self.orb_type == "f":  # f orbital [7F]
            return 7
        else:
            raise NotImplementedError(
                "The atomic orbital type '" + self.orb_type + "' is not implemented."
            )

    def __calculate_coefficients(self, index: int, orb_type: str) -> float:
        if self.orb_type == "s":  # s orbital
            return self.coefficients[index][1] * pow(
                2.0 * self.alpha[index] / math.pi, 0.75
            )
        elif self.orb_type == "p":  # p orbital
            return self.coefficients[index][1] * pow(
                128.0 * pow(self.alpha[index], 5) / pow(math.pi, 3), 0.25
            )
        elif self.orb_type == "d":  # d orbital
            return self.coefficients[index][1] * pow(
                2048.0 * pow(self.alpha[index], 7) / pow(math.pi, 3), 0.25
            )
        elif self.orb_type == "f":  # f orbital
            return self.coefficients[index][1] * pow(
                32768.0 * pow(self.alpha[index], 9) / pow(math.pi, 3), 0.25
            )
        else:
            raise NotImplementedError(
                "The atomic orbital type '" + orb_type + "' is not implemented."
            )


class Atom:
    """
    Provide Atom.
    """

    def __init__(self, name: str, elem: int, x: float, y: float, z: float):
        self.name = name
        self.elem = elem
        self.coordinates = times_angstrom_per_bohr([x, y, z])
        self.gaussian_orbitals: List[GaussianOrbital] = list()
        self.min_alpha = 0
        self.sum_chi_step = 0  # sum_chi_step will be used in case atom will be skipped

    @classmethod
    def from_molden_line(cls, line: str) -> Any:
        split = line.split()
        parsed_line = (
            # example : C 1 6 6.1052194517 1.7410254664 1.1040800477
            split[0],
            # ignore split[1], which is the number of the atom
            int(split[2]),
            float(split[3]),
            float(split[4]),
            float(split[5]),
        )
        return cls(*parsed_line)


class ElectronicData:
    """
    Provide electronic data.
    """

    def __init__(self, atoms: List[Atom], mo: List[MolecularOrbital]) -> None:
        self.atoms = atoms
        self.mo = mo

    def get_occupations(self):
        occupations = []
        for mo in self.mo:
            occupations.append(mo.occupation)
        return occupations

    def get_orbital_energies(self):
        energies = []
        for mo in self.mo:
            energies.append(mo.energy)
        return energies



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
"""
Provides the HapticPointer class.
"""
from typing import Any

from vtk import (
    vtkTrivialProducer,
    vtkAlgorithmOutput,
)
from vtkmodules.vtkCommonCore import vtkPoints
from vtkmodules.vtkCommonDataModel import vtkPolyData


class HapticPointerData:
    """
    Displays haptic pointer in a 3D view.
    """

    def __init__(self) -> None:

        # Center only
        self.__center = vtkPoints()
        self.__center.InsertPoint(0, 0.0, 0.0, 0.0)
        self.__center_data = vtkPolyData()
        self.__center_data.SetPoints(self.__center)
        self.__center_data_producer = vtkTrivialProducer()
        self.__center_data_producer.SetOutput(self.__center_data)

    @property
    def position(self) -> Any:
        return self.__center.GetPoint(0)

    @property
    def output(self) -> vtkAlgorithmOutput:
        return self.__center_data_producer.GetOutputPort()

    def get_center_data(self) -> Any:
        return self.__center_data

    def update_pointer_position(self, pos: Any) -> None:
        """
        Update pointer position.
        """
        self.__center.InsertPoint(0, pos.x, pos.y, pos.z)
        self.__center_data_producer.Modified()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__copyright__ = """ This code is licensed under the 3-clause BSD license.
Copyright ETH Zurich, Department of Chemistry and Applied Biosciences, Reiher Group.
See LICENSE.txt for details.
"""
from typing import Optional
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D


class EnergyDiagram(object):
    """
    This class has been derived and adapted from https://github.com/giacomomarchioro/PyEnergyDiagrams.
    """

    # --- Energy profile diagram---
    # This is a simple script to plot energy profile diagram using matplotlib.
    # autopep8: off
    # E|          4__       # noqa: W605,W1401
    # n|   2__    /  \      # noqa: W605,W1401
    # e|1__/  \__/5   \     # noqa: W605,W1401
    # r|  3\__/       6\__  # noqa: W605,W1401
    # g|
    # y|
    # autopep8: on

    def __init__(self, aspect='equal'):
        # plot parameters
        self.ratio = 1.6181
        self.dimension = 'auto'
        self.space = 'auto'
        self.offset = 'auto'
        self.offset_ratio = 0.02
        self.color_bottom_text = 'black'
        self.aspect = aspect
        self.round_energies_at_digit = "keep all digits"
        self.top_text_fontsize = "medium"
        self.bottom_text_fontsize = "medium"
        self.right_text_fontsize = "medium"
        self.left_text_fontsize = "medium"
        # data
        self.pos_number = 0
        self.energies = []
        self.positions = []
        self.colors = []
        self.alphas = []
        self.top_texts = []
        self.bottom_texts = []
        self.left_texts = []
        self.right_texts = []
        self.links = []
        self.arrows = []
        self.electons_boxes = []
        self.level_linestyles = []
        self.level_linewidths = []
        # matplotlib figure handlers
        self.fig = None
        self.ax = None

    def add_level(self, energy, bottom_text='', position=None, color='k', alpha=1.0,
                  top_text='Energy', right_text='', left_text='', linestyle='solid',
                  linewidth=1):
        """
        Method of ED class
        This method add a new energy level to the plot.

        Parameters
        ----------
        energy : int
                 The energy of the level in Kcal mol-1
        bottom_text  : str
                The text on the bottom of the level (label of the level)
                (default '')
        position  : str
                The position of the level in the plot. Keep it empty to add
                the level on the right of the previous level use 'last' as
                argument for adding the level to the last position used
                for the level before.
                An integer can be used for adding the level to an arbitrary
                position.
                (default  None)
        color  : str
                Color of the level  (default  'k')
        alpha  : float
                Transparency of the level (default  1.0)
        top_text  : str
                Text on the top of the level. By default it will print the
                energy of the level. (default  'Energy')
        right_text  : str
                Text at the right of the level. (default  '')
        left_text  : str
                Text at the left of the level. (default  '')
        linestyle  : str
                The linestyle of the level, one of the following values:
                'solid', 'dashed', 'dashdot', 'dotted' (default  'solid')




        Returns
        -------
        Append to the class data all the information regarding the level added
        """

        if position is None:
            position = self.pos_number + 1
            self.pos_number += 1
        elif isinstance(position, (int, float)):
            pass
        elif position == 'last' or position == 'l':
            position = self.pos_number
        else:
            raise ValueError(
                "Position must be None or 'last' (abr. 'l') or in case an integer or float specifing the position. " +
                "It was: %s" % position)
        if top_text == 'Energy':
            if self.round_energies_at_digit == "keep all digits":
                top_text = energy
            else:
                top_text = round(energy, self.round_energies_at_digit)

        link = []
        self.colors.append(color)
        self.alphas.append(alpha)
        self.energies.append(energy)
        self.positions.append(position)
        self.top_texts.append(top_text)
        self.bottom_texts.append(bottom_text)
        self.left_texts.append(left_text)
        self.right_texts.append(right_text)
        self.links.append(link)
        self.level_linestyles.append(linestyle)
        self.level_linewidths.append(linewidth)
        self.arrows.append([])

    def add_arrow(self, start_level_id, end_level_id):
        """
        Method of ED class
        Add a arrow between two energy levels using IDs of the level. Use
        self.plot(show_index=True) to show the IDs of the levels.

        Parameters
        ----------
        start_level_id : int
                 Starting level ID
        end_level_id : int
                 Ending level ID

        Returns
        -------
        Append arrow to self.arrows

        """
        self.arrows[start_level_id].append(end_level_id)

    def add_link(self, start_level_id, end_level_id,
                 color='k',
                 ls='--',
                 linewidth=1,
                 alpha=1.0
                 ):
        """
        Method of ED class
        Add a link between two energy levels using IDs of the level. Use
        self.plot(show_index=True) to show the IDs of the levels.

        Parameters
        ----------
        start_level_id : int
                 Starting level ID
        end_level_id : int
                 Ending level ID
        color : str
                color of the line
        ls : str
                line styple e.g. -- , ..
        linewidth : int
                line width
        alpha : float
                alpha

        Returns
        -------
        Append link to self.links

        """
        self.links[start_level_id].append((end_level_id, ls, linewidth, color, alpha))

    def plot(self, show_IDs=False, ylabel="Electronic Energy / kJ mol$^{-1}$", ax: Optional[plt.Axes] = None,
             ylimits=None):
        """
        Method of ED class
        Plot the energy diagram. Use show_IDs=True for showing the IDs of the
        energy levels and allowing an easy linking.

        Parameters
        ----------
        show_IDs : bool
            show the IDs of the energy levels
        ylabel : str
            The label to use on the left-side axis. "El. Energy / $kJ$ $mol^{-1}$" by default.
        ax : plt.Axes
            The axes to plot onto. If not specified, a Figure and Axes will be
            created for you.

        Returns
        -------
        fig (plt.figure) and ax (fig.add_subplot())
        """

        # Create a figure and axis if the user didn't specify them.
        if not ax:
            self.fig = plt.figure()
            self.ax = self.fig.add_subplot(111, aspect=self.aspect)
        # Otherwise register the axes and figure the user passed.
        else:
            self.ax = ax
            self.fig = ax.figure

            # Constrain the target axis to have the proper aspect ratio
            self.ax.set_aspect(self.aspect)

        self.ax.set_ylabel(ylabel)
        # self.ax.axes.get_xaxis().set_visible(False)
        self.ax.set_xticks([])
        self.ax.spines['top'].set_visible(False)
        self.ax.spines['right'].set_visible(False)
        self.ax.spines['bottom'].set_visible(False)

        self.__auto_adjust()

        data = list(zip(self.energies,  # 0
                        self.positions,  # 1
                        self.bottom_texts,  # 2
                        self.top_texts,  # 3
                        self.colors,  # 4
                        self.right_texts,  # 5
                        self.left_texts,  # 6
                        self.level_linestyles,  # 7
                        self.level_linewidths,  # 8
                        self.alphas)  # 9
                    )
        for level in data:
            start = level[1] * (self.dimension + self.space)
            self.ax.plot([start, start + self.dimension], [level[0], level[0]],
                         color=level[4],
                         alpha=level[9],
                         linestyle=level[7],
                         linewidth=level[8],
                         solid_capstyle='butt')
            self.ax.text(start + self.dimension / 2.,  # X
                         level[0] + self.offset,  # Y
                         level[3],  # self.top_texts
                         horizontalalignment='center',
                         verticalalignment='bottom',
                         fontsize=self.top_text_fontsize)

            self.ax.text(start + self.dimension,  # X
                         level[0],  # Y
                         level[5],  # self.right_text
                         horizontalalignment='left',
                         verticalalignment='center',
                         color=self.color_bottom_text,
                         fontsize=self.left_text_fontsize)

            self.ax.text(start,  # X
                         level[0],  # Y
                         level[6],  # self.left_text
                         horizontalalignment='right',
                         verticalalignment='center',
                         color=self.color_bottom_text,
                         fontsize=self.right_text_fontsize)

            self.ax.text(start + self.dimension / 2.,  # X
                         level[0] - self.offset * 1.5,  # Y
                         level[2],  # self.bottom_text
                         horizontalalignment='center',
                         verticalalignment='top',
                         color=self.color_bottom_text,
                         fontsize=self.bottom_text_fontsize)
        if show_IDs:
            # for showing the ID allowing the user to identify the level
            for ind, level in enumerate(data):
                start = level[1] * (self.dimension + self.space)
                self.ax.text(start, level[0] + self.offset, str(ind),
                             horizontalalignment='right', color='red')

        for idx, arrow in enumerate(self.arrows):
            # by Kalyan Jyoti Kalita: put arrows between to levels
            # x1, x2   y1, y2
            for i in arrow:
                start = self.positions[idx] * (self.dimension + self.space)
                x1 = start + 0.5 * self.dimension
                x2 = start + 0.5 * self.dimension
                y1 = self.energies[idx]
                y2 = self.energies[i]
                gap = y1 - y2
                gapnew = '{0:.2f}'.format(gap)
                middle = y1 - 0.5 * gap  # warning: this way works for negative HOMO/LUMO energies
                self.ax.annotate("", xy=(x1, y1), xytext=(x2, middle), arrowprops=dict(
                    color='green', width=2.5, headwidth=5))
                self.ax.annotate(gapnew, xy=(x2, y2), xytext=(x1, middle), color='green',
                                 arrowprops=dict(width=2.5, headwidth=5, color='green'),
                                 bbox=dict(boxstyle='round', fc='white'),
                                 ha='center', va='center')

        for idx, link in enumerate(self.links):
            # here we connect the levels with the links
            # x1, x2   y1, y2
            for i in link:
                # i is a tuple: (end_level_id,ls,linewidth,color,alpha)
                start = self.positions[idx] * (self.dimension + self.space)
                x1 = start + self.dimension
                x2 = self.positions[i[0]] * (self.dimension + self.space)
                y1 = self.energies[idx]
                y2 = self.energies[i[0]]
                line = Line2D([x1, x2], [y1, y2],
                              ls=i[1],
                              linewidth=i[2],
                              color=i[3],
                              alpha=i[4])
                self.ax.add_line(line)

        if ylimits is not None:
            self.ax.set_ylim(ylimits)
        else:
            self.ax.xaxis.set_label_coords(0.5, -0.2)

    def __auto_adjust(self):
        """
        Method of ED class
        This method use the ratio to set the best dimension and space between
        the levels.

        Affects
        -------
        self.dimension
        self.space
        self.offset

        """
        # Max range between the energy
        energy_variation = abs(max(self.energies) - min(self.energies))
        if self.dimension == 'auto' or self.space == 'auto':
            # Unique positions of the levels
            unique_positions = float(len(set(self.positions)))
            space_for_level = energy_variation * self.ratio / unique_positions
            self.dimension = space_for_level * 0.7
            self.space = space_for_level * 0.3

        if self.offset == 'auto':
            self.offset = energy_variation * self.offset_ratio
